#ifndef _WANPIPE_LAPB_H
#define _WANPIPE_LAPB_H 

#include <linux/wanpipe_cfg.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_snmp.h>
#include <linux/wanrouter.h>
#include <linux/wanproc.h>
#include <linux/wanpipe_lapb_kernel.h>

#define	LAPB_HEADER_LEN	20		/* LAPB over Ethernet + a bit more */

#define	LAPB_ACK_PENDING_CONDITION	0x01
#define	LAPB_REJECT_CONDITION		0x02
#define	LAPB_PEER_RX_BUSY_CONDITION	0x04
#define LAPB_RNR_SENT_CONDITION		0x08

#define LAPB_RX_Q_H_LIMIT 4
#define LAPB_RX_Q_L_LIMIT 3

	/* Control field templates */
#define	LAPB_I		0x00	/* Information frames */
#define	LAPB_S		0x01	/* Supervisory frames */
#define	LAPB_U		0x03	/* Unnumbered frames */

#define	LAPB_RR		0x01	/* Receiver ready */
#define	LAPB_RNR	0x05	/* Receiver not ready */
#define	LAPB_REJ	0x09	/* Reject */

#define	LAPB_SABM	0x2F	/* Set Asynchronous Balanced Mode */
#define	LAPB_SABME	0x6F	/* Set Asynchronous Balanced Mode Extended */
#define	LAPB_DISC	0x43	/* Disconnect */
#define	LAPB_DM		0x0F	/* Disconnected mode */
#define	LAPB_UA		0x63	/* Unnumbered acknowledge */
#define	LAPB_FRMR	0x87	/* Frame reject */

#define LAPB_ILLEGAL	0x100	/* Impossible to be a real frame type */
#define LAPB_CORRUPT    0x101   /* Corrupted address field */

#define	LAPB_SPF	0x10	/* Poll/final bit for standard LAPB */
#define	LAPB_EPF	0x01	/* Poll/final bit for extended LAPB */

#define	LAPB_FRMR_W	0x01	/* Control field invalid	*/
#define	LAPB_FRMR_X	0x02	/* I field invalid		*/
#define	LAPB_FRMR_Y	0x04	/* I field too long		*/
#define	LAPB_FRMR_Z	0x08	/* Invalid N(R)			*/

#define	LAPB_POLLOFF	0
#define	LAPB_POLLON	1

	/* LAPB C-bit */
#define LAPB_COMMAND	1
#define LAPB_RESPONSE	2

#define	LAPB_ADDR_A	0x03
#define	LAPB_ADDR_B	0x01
#define	LAPB_ADDR_C	0x0F
#define	LAPB_ADDR_D	0x07

#define TX_TIMEOUT  5
	/* Define Link State constants. */
	enum {
		LAPB_STATE_DISCONNECTED,		/* Disconnected State		*/
		LAPB_STATE_CONNECTING,		/* Awaiting Connection State	*/
		LAPB_STATE_DISCONNECTING, 	/* Awaiting Disconnection State	*/
		LAPB_STATE_CONNECTED,		/* Data Transfer State		*/
		LAPB_STATE_FRAME_REJECT		/* Frame Reject State		*/
	};

#define	LAPB_DEFAULT_MODE		(LAPB_STANDARD | LAPB_SLP | LAPB_DTE)
#define	LAPB_DEFAULT_WINDOW		7		/* Window=7 */
#define	LAPB_DEFAULT_T1			(5 * HZ)	/* T1=5s    */
#define	LAPB_DEFAULT_T2			(1 * HZ)	/* T2=1s    */
#define	LAPB_DEFAULT_N2			20		/* N2=20    */

#define	LAPB_SMODULUS	8
#define	LAPB_EMODULUS	128

#define LAPB_UNREG_CRIT		0
#define LAPB_SMP_CRIT		1
#define LAPB_AWAITING_KICK 	2
#define LAPB_FRAME_SENT		3
#define LAPB_RX_SMP		4
#define LAPB_BH_RUNNING		5
#define LAPB_TIMER_CRIT		6
#define LAPB_REQUEUING		7
#define LAPB_DOWN		8

/*
 *	Information about the current frame.
 */
struct lapb_frame {
	unsigned short		type;		/* Parsed type		*/
	unsigned short		nr, ns;		/* N(R), N(S)		*/
	unsigned char		cr;		/* Command/Response	*/
	unsigned char		pf;		/* Poll/Final		*/
	unsigned char		control[2];	/* Original control data*/
};

struct lapb_protocol_stats {
	unsigned int rr_tx;
	unsigned int rnr_tx;
	unsigned int rej_tx;
	unsigned int sabm_tx;
	unsigned int ua_tx;
	unsigned int disc_tx;
	unsigned int dm_tx;
	unsigned int ifrm_tx;
	unsigned int frmr_tx;
	unsigned int ill_tx;
	
	unsigned int rr_rx;
	unsigned int rnr_rx;
	unsigned int rej_rx;
	unsigned int sabm_rx;
	unsigned int ua_rx;
	unsigned int disc_rx;
	unsigned int dm_rx;
	unsigned int ifrm_rx;
	unsigned int frmr_rx;
	unsigned int ill_rx;
};

/*
 *	The per LAPB connection control structure.
 */
typedef struct lapb_cb {

	struct net_device 	*next;
	
	unsigned long		critical;
	
	/* Link status fields */
	unsigned int		mode;
	unsigned char		state;
	volatile unsigned short	vs, vr, va;
	unsigned char		condition;
	unsigned short		n2, n2count, n4count;
	unsigned int		t1, t2, t3, t3_orig, t4, t4_orig;
	struct timer_list	t1timer, t2timer, t4timer;

	/* Internal control information */
	struct sk_buff_head	write_queue;
	struct sk_buff_head	ack_queue;
	struct sk_buff_head	async_queue;
	struct sk_buff_head	rx_queue;
	unsigned char		window;
	unsigned int	 	mtu;
	
	/* FRMR control information */
	struct lapb_frame	frmr_data;
	unsigned char		frmr_type;

	/* Transmit device */
	struct net_device	*dev;
	struct net_device 	*tx_dev;
	unsigned char 		*hw_name;

	/* X25 layer interface */
	void * x25_link;
	void * sk_id;
	
	spinlock_t tx_lock;
	spinlock_t bh_lock;

	struct proc_dir_entry* dent;
	struct tasklet_struct lapb_task;

	struct net_device_stats ifstats;
	struct lapb_protocol_stats lapbstats;

	unsigned int kick_cnt;
	unsigned char label[WAN_IF_LABEL_SZ+1];

	/* X25_SW */
	unsigned char virtual_addr[WAN_ADDRESS_SZ+1];
	unsigned char real_addr[WAN_ADDRESS_SZ+1];

	unsigned short old_nr,old_nr_type,nr_type;
} lapb_cb;

/* lapb_iface.c */
extern void lapb_connect_confirmation(lapb_cb *, int);
extern void lapb_connect_indication(lapb_cb *, int);
extern void lapb_disconnect_confirmation(lapb_cb *, int);
extern void lapb_disconnect_indication(lapb_cb *, int);
extern int  lapb_data_indication(lapb_cb *, struct sk_buff *);
extern int  lapb_data_transmit(lapb_cb *, struct sk_buff *);

/* lapb_proc.c */
extern int lapb_proc_init (void);
extern void lapb_proc_cleanup (void);
extern int lapb_proc_add (lapb_cb* lapb);
extern int lapb_proc_delete (lapb_cb* lapb);
extern int lapb_proc_check_use(lapb_cb* lapb);


/* lapb_in.c */
extern void lapb_data_input(lapb_cb *, struct sk_buff *);

/* lapb_out.c */
extern void lapb_kick(unsigned long data);
extern int lapb_transmit_buffer(lapb_cb *, struct sk_buff *, int,char);
extern void lapb_establish_data_link(lapb_cb *);
extern void lapb_enquiry_response(lapb_cb *);
extern void lapb_timeout_response(lapb_cb *);
extern  void lapb_check_iframes_acked(lapb_cb *, unsigned short);
extern void lapb_check_need_response(lapb_cb *, int, int);

/* lapb_subr.c */
extern void lapb_clear_queues(lapb_cb *);
extern void lapb_frames_acked(lapb_cb *, unsigned short);
extern void lapb_requeue_frames(lapb_cb *);
extern  int  lapb_validate_nr(lapb_cb *, unsigned short);
extern void lapb_decode(lapb_cb *, struct sk_buff *, struct lapb_frame *);
extern void lapb_send_control(lapb_cb *, int, int, int);
extern void lapb_transmit_frmr(lapb_cb *);
extern void lapb_update_stats (lapb_cb *lapb, int type, int len, char direction);
extern void lapb_state_change(lapb_cb *lapb, int state);
extern int lapb_trigger_kick(lapb_cb *lapb,char kick);
extern void lapb_reconnect(lapb_cb*lapb);

/* lapb_timer.c */
extern  int lapb_queue_tq (struct tq_struct *bh_pointer);
extern  int lapb_tq_mark_bh (void);
extern void lapb_start_t1timer(lapb_cb *);
extern void lapb_start_t2timer(lapb_cb *);
extern void lapb_start_t4timer(lapb_cb *, unsigned char rest_t4_count);
extern void lapb_stop_t1timer(lapb_cb *);
extern void lapb_stop_t2timer(lapb_cb *);
extern void lapb_stop_t4timer(lapb_cb *);
extern int  lapb_t1timer_running(lapb_cb *);

/*
 * Debug levels.
 *	0 = Off
 *	1 = State Changes
 *	2 = Packets I/O and State Changes
 *	3 = Hex dumps, Packets I/O and State Changes.
 */
//#define	LAPB_DEBUG	0

extern atomic_t lapb_skb_alloc;
extern atomic_t lapb_mem_alloc;

#define LAPB_DEBUG_MEM
#ifdef LAPB_DEBUG_MEM

 #define LAPB_SKB_INC(x)  atomic_add(x,&lapb_skb_alloc)
 #define LAPB_SKB_DEC(x)  atomic_sub(x,&lapb_skb_alloc)

 #define ALLOC_SKB(skb,len) { skb = dev_alloc_skb(len);			\
			      if (skb != NULL){ LAPB_SKB_INC(skb->truesize);} else {WAN_MEM_ASSERT("LAPB"); }}
 #define KFREE_SKB(skb)     { LAPB_SKB_DEC(skb->truesize); dev_kfree_skb_any(skb); }

 #define LAPB_MEM_INC(x)  atomic_add(x,&lapb_mem_alloc)
 #define LAPB_MEM_DEC(x)  atomic_sub(x,&lapb_mem_alloc)

 #define KMALLOC(ptr,len,flag)	{ ptr=kmalloc(len, flag); \
	  			  if (ptr != NULL){ LAPB_MEM_INC(len);} else {WAN_MEM_ASSERT("LAPB"); }}
 #define KFREE(ptr)		{LAPB_MEM_DEC(sizeof(*ptr)); kfree(ptr);}
			  
			      
#else
 #define LAPB_SKB_INC(x)
 #define LAPB_SKB_DEC(x)			      
 #define LAPB_MEM_INC(x)
 #define LAPB_MEM_DEC(x)
 
 #define KMALLOC(ptr,len,flag)	ptr=kmalloc(len, flag)
 #define KFREE(ptr)		kfree(ptr)				      
			
 #define ALLOC_SKB(new_skb,len, dsp) new_skb = dev_alloc_skb(len)
 #define KFREE_SKB(skb)	             dev_kfree_skb_any(skb)
#endif


#define LAPB_DEBUG 0



#endif
