/*****************************************************************************
* lapb_iface.c	WANPIPE LapB Protocol Module. Main code.
*
* Author:	Nenad Corbic
*
* Copyright:	(c)2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*	This module:
*		This module is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
*/

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include <linux/wanpipe_kernel.h>
#include "wanpipe_lapb.h"

static int lapb_if_send (struct sk_buff*, struct net_device*);

struct lapb_x25_register_struct x25_protocol;

struct net_device *volatile lapb_list;
atomic_t lapb_skb_alloc;
atomic_t lapb_mem_alloc;
static spinlock_t lapb_spin_lock;

static void lapb_link_up (struct net_device *dev);
void lapb_link_down (struct net_device *dev);



lapb_cb* dev_to_lapb(struct net_device *dev)
{
	return (lapb_cb*)dev->priv;
}


/*
 *	Free an allocated lapb control block. This is done to centralise
 *	the MOD count code.
 */
static void lapb_free_dev(struct net_device *dev)
{
	lapb_cb *lapb_cb;
	
	if ((lapb_cb=dev->priv) != NULL){
		dev->priv=NULL;
		KFREE(lapb_cb);
	}

	if (wan_atomic_read(&dev->refcnt)){
		DEBUG_EVENT("%s: Dev=%s Major error dev is still in use %d\n",
				__FUNCTION__,
				wan_netif_name(dev),
				wan_atomic_read(&dev->refcnt));
	}

	wan_netif_free(dev);

#ifndef LINUX_2_6
	MOD_DEC_USE_COUNT;
#endif
}

/*
 *	Socket removal during an interrupt is now safe.
 */
static void lapb_remove_cb(struct net_device *dev)
{
	struct net_device *cur_dev;
	struct net_device *prev_dev;
	unsigned long flags;

	spin_lock_irqsave(&lapb_spin_lock,flags);

	if ((cur_dev = lapb_list) == dev) {
		lapb_list = *(struct net_device**)cur_dev->priv;
		spin_unlock_irqrestore(&lapb_spin_lock,flags);


		unregister_netdev(dev);
		return;
	}

	while (cur_dev != NULL && *(struct net_device**)cur_dev->priv != NULL) {
		prev_dev=cur_dev;
		cur_dev = *(struct net_device**)cur_dev->priv;
		if (cur_dev == dev) {
			*(struct net_device**)prev_dev->priv = *(struct net_device**)dev->priv;
			spin_unlock_irqrestore(&lapb_spin_lock,flags);
		
			unregister_netdev(dev);
			return;
		}
	}

	spin_unlock_irqrestore(&lapb_spin_lock,flags);
}

/*
 *	Add a socket to the bound sockets list.
 */
static void lapb_insert_cb(struct net_device *dev)
{
	unsigned long flags;

	spin_lock_irqsave(&lapb_spin_lock,flags);

	*(struct net_device **)dev->priv = lapb_list;
	lapb_list = dev;

	spin_unlock_irqrestore(&lapb_spin_lock,flags);
}

/*
 *	Convert the integer token used by the device driver into a pointer
 *	to a LAPB control structure.
 */
static struct net_device *master_to_lapb_dev(struct net_device *dev)
{
	struct net_device *tmp_dev;
	unsigned long flags;

	spin_lock_irqsave(&lapb_spin_lock,flags);
	
	for (tmp_dev = lapb_list; tmp_dev != NULL; tmp_dev= *(struct net_device **)tmp_dev->priv){
		lapb_cb *lapb = (lapb_cb *)tmp_dev->priv;
		if (lapb->tx_dev == dev){
			spin_unlock_irqrestore(&lapb_spin_lock,flags);	
			return tmp_dev;
		}
	}

	spin_unlock_irqrestore(&lapb_spin_lock,flags);
	return NULL;
}

/*
 *	Create an empty LAPB control block.
 */
static lapb_cb *lapb_create_cb(void)
{
	lapb_cb *lapb;

	KMALLOC(lapb,sizeof(*lapb), GFP_KERNEL);
	if (lapb == NULL)
		return NULL;

	memset(lapb, 0x00, sizeof(lapb_cb));

	skb_queue_head_init(&lapb->write_queue);
	skb_queue_head_init(&lapb->ack_queue);
	skb_queue_head_init(&lapb->async_queue);
	skb_queue_head_init(&lapb->rx_queue);


	init_timer(&lapb->t1timer);
	init_timer(&lapb->t2timer);
	init_timer(&lapb->t4timer);

	tasklet_init(&lapb->lapb_task,lapb_kick,(unsigned long)lapb);
#if 0
	lapb->lapb_task.sync = 0;
	lapb->lapb_task.routine = (void *)(void *)lapb_kick;
	lapb->lapb_task.data = lapb;
#endif	
	spin_lock_init(&lapb->tx_lock);
	spin_lock_init(&lapb->bh_lock);

	lapb->t1      = LAPB_DEFAULT_T1;
	lapb->t2      = LAPB_DEFAULT_T2;
	lapb->n2      = LAPB_DEFAULT_N2;
	lapb->mode    = LAPB_DEFAULT_MODE;
	lapb->window  = LAPB_DEFAULT_WINDOW;
	lapb->state   = LAPB_STATE_DISCONNECTED;

	return lapb;
}

static struct net_device *lapb_create_dev(char *dev_name)
{
	struct net_device *dev;
	int iftype = WAN_IFT_OTHER;
	int err;

	dev=wan_netif_alloc(dev_name,iftype,&err);
	if (dev == NULL){
		return NULL;
	}
	wan_atomic_set(&dev->refcnt,0);
	
#ifndef LINUX_2_6
	MOD_INC_USE_COUNT;
#endif
	
	return dev;
}

static void if_tx_timeout (struct net_device *dev)
{
	/* If our device stays busy for at least 5 seconds then we will
	 * kick start the device by making dev->tbusy = 0.  We expect
	 * that our device never stays busy more than 5 seconds. So this                 
	 * is only used as a last resort.
	 */

	printk (KERN_INFO "%s: Transmit timed out.\n", 
			dev->name);
	netif_wake_queue (dev);

}

int lapb_ioctl (struct net_device *dev, struct ifreq *ifr, int cmd)
{
	lapb_cb *lapb;
	int err = -EOPNOTSUPP;
	
	if (!dev || !(dev->flags&IFF_UP))
		return -ENODEV;
	
	if ((lapb=dev->priv) == NULL)
		return -ENODEV;
		
	switch (cmd){

	case SIOCS_LAPB_NEW_X25_CONF:
		/* Register and configure a new X25 link */
		printk(KERN_INFO "%s: LAPB IOCTL  NEW X25\n",dev->name); 
		break;

	case SIOCS_LAPB_DEL_X25_CONF:
		/* Unregister an X25 link */
		printk(KERN_INFO "%s: LAPB IOCTL  DEL X25\n",dev->name); 
		break;

	case SIOCG_LAPB_CONF: 
		/* Get Lapb Configuration */
		printk(KERN_INFO "%s: LAPB IOCTL  GET LAPB CONF\n",dev->name); 
		break;

	case SIOCG_LAPB_STATUS:
		/* Get Lapb Status */
		err=lapb->state == LAPB_STATE_CONNECTED?1:0;
		break;
		
	default:
		err = -EOPNOTSUPP;
	}
	
	return err;
	
}

int lapb_set_config (struct net_device *dev,struct ifmap *map)
{
	return 0;	
}

static struct net_device_stats gstats;
static struct net_device_stats* lapb_ifstats (struct net_device *dev)
{
	lapb_cb *lapb;
	
	if (!dev || !(dev->flags&IFF_UP))
		return &gstats;
	
	if ((lapb=dev->priv) == NULL)
		return &gstats;

	return &lapb->ifstats;
}


static int if_dev_open(struct net_device *dev)
{
	struct in_ifaddr *ifaddr;
	struct in_device *in_dev;
	lapb_cb *lapb;

	if (!(lapb=dev->priv)){
		return -EINVAL;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
	if (netif_running(dev))
		return -EBUSY;
#endif

	netif_start_queue(dev);

	if ((in_dev = in_dev_get(dev)) == NULL){
		/* Used by API thus remve if_tx_timeout */
		goto if_dev_open_api_mode;
	}

	if ((ifaddr = in_dev->ifa_list)== NULL ){
		goto if_dev_open_api_mode;
	}

	if (lapb->tx_dev && lapb->tx_dev->priv &&
	    ((wanpipe_common_t*)lapb->tx_dev->priv)->state == WANSOCK_CONNECTED){
		lapb_link_up(dev);	
	}
	
	return 0;

if_dev_open_api_mode:
	dev->tx_timeout = NULL;

	if (lapb->tx_dev && lapb->tx_dev->priv &&
	    ((wanpipe_common_t*)lapb->tx_dev->priv)->state == WANSOCK_CONNECTED){
		lapb_link_up(dev);	
	}
	return 0;

}


static int if_dev_close(struct net_device *dev)
{
	lapb_cb *lapb = dev_to_lapb(dev);

	netif_stop_queue(dev); 
	lapb_link_down(dev);

	if (lapb){
		lapb_send_control(lapb, LAPB_DM, 0, LAPB_COMMAND);
	}
	return 0;	
}

static int lapb_if_init(struct net_device *dev)
{
	dev->open		= &if_dev_open;
	dev->stop		= &if_dev_close;
	dev->hard_header	= NULL;
	dev->rebuild_header	= NULL;
	dev->hard_start_xmit	= &lapb_if_send;
	dev->get_stats		= &lapb_ifstats;
	dev->tx_timeout		= &if_tx_timeout;
	dev->watchdog_timeo	= TX_TIMEOUT;
	dev->hard_header_len	= 6;
	dev->do_ioctl		= &lapb_ioctl;
	dev->set_config		= &lapb_set_config;
	
		/* Initialize media-specific parameters */
	dev->flags		|= IFF_POINTOPOINT;
	dev->flags		|= IFF_NOARP;

	dev->tx_queue_len = 10;
  	dev->mtu = 4096;
	dev->type = ARPHRD_LAPB;

	/* Initialize socket buffers */
	dev_init_buffers(dev);
	return 0;
}


int lapb_register(struct net_device *device, char *dev_name, char *hw_name, struct net_device **new_dev)
{
	lapb_cb *lapb;
	struct net_device *dev;
	int err;

	WAN_ASSERT2((!dev_name),-ENODEV);
	WAN_ASSERT2((!hw_name),-ENODEV);
	WAN_ASSERT2((!device),-ENODEV);
	
	if (master_to_lapb_dev(device) != NULL){
		printk(KERN_INFO "LAPB: Device already registerd for %s\n",
				device->name);
		return -EEXIST;
	}

	if (!new_dev){
		printk(KERN_INFO "LAPB: Failed to create new lapb if: no new_dev ptr!\n");
		return -EFAULT;
	}
	
	if ((lapb = lapb_create_cb()) == NULL){
		printk(KERN_INFO "LAPB: Failed to create cb %s\n",
				device->name);
		return -ENOMEM;
	}

	if ((dev = lapb_create_dev(dev_name)) == NULL){
		printk(KERN_INFO "LAPB: Failed to create dev %s\n",
				dev_name);
		KFREE(lapb);
		return -ENOMEM;
	}

	dev->init = &lapb_if_init;
	wan_netif_set_priv(dev, lapb);
	
	/* Register x25 link */
	if (!lapb->x25_link && IS_FUNC_CALL(x25_protocol,x25_register)){
		err=lapb_x25_register(dev,NULL,NULL);
		if (lapb->x25_link == NULL){
			KFREE(lapb);
			lapb_free_dev(dev);
			printk(KERN_INFO "%s: Lapb register: failed to register x25_link\n",
					dev_name);
			return err;
		}
	}

	rtnl_lock();
	if ((err=register_netdevice(dev)) != 0){
		rtnl_unlock();
		KFREE(lapb);
		lapb_free_dev(dev);
		printk(KERN_INFO "LAPB: Failed to register net dev%s\n",
				device->name);
		return err;
	}
	rtnl_unlock();
	
	lapb->tx_dev = device;
	lapb->hw_name = hw_name;
	lapb->dev = dev;
	lapb_insert_cb(dev);
	lapb_start_t1timer((lapb_cb*)dev->priv);

#ifdef CONFIG_PROC_FS 
	lapb_proc_add(lapb);
#endif
		
	*new_dev=dev;
	return 0;
}

int lapb_unregister(struct net_device *dev)
{
	lapb_cb *lapb;
	int err;

	WAN_ASSERT2((!dev),-ENODEV);	
	
	if (dev->flags & IFF_UP){
		printk(KERN_INFO "%s: Failed to unregister: Device UP!\n",
				dev->name);
		return -EBUSY;
	}
	
	lapb = dev_to_lapb(dev);
	if (!lapb){
		printk(KERN_INFO "%s: Failed to unregister: No lapb dev!\n",
				dev->name);
		return -ENODEV;
	}

	if (lapb_proc_check_use(lapb)){
		printk(KERN_INFO "%s: Error: Lapb proc directory in use!\n",
				dev->name);
		return -EBUSY;
	}
	
	set_bit (LAPB_UNREG_CRIT, &lapb->critical);
	tasklet_kill(&lapb->lapb_task);
	schedule();
	
	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_unregister_link)){
			err = x25_protocol.x25_unregister_link(lapb->x25_link);
			if (err != 0){
				clear_bit (LAPB_UNREG_CRIT, &lapb->critical);
				tasklet_init(&lapb->lapb_task,lapb_kick,(unsigned long)lapb);
				lapb_trigger_kick(lapb,1);
				return err;
			}
		}
	}

	printk(KERN_INFO "%s: Unregistering Lapb\n",dev->name);

#ifdef CONFIG_PROC_FS
	lapb_proc_delete(lapb);
#endif
	
	lapb_stop_t1timer(lapb);
	lapb_stop_t2timer(lapb);
	lapb_stop_t4timer(lapb);

	lapb_clear_queues(lapb);

	lapb_remove_cb(dev);
	lapb_free_dev(dev);

	clear_bit (LAPB_UNREG_CRIT, &lapb->critical);
	
	printk(KERN_INFO "%s: Unregistering Lapb OK\n",dev->name);

	return LAPB_OK;
}


int lapb_getparms(struct net_device *dev, struct lapb_parms_struct *parms)
{
	lapb_cb *lapb;

	WAN_ASSERT2((!dev),LAPB_BADTOKEN);
	WAN_ASSERT2((!parms),LAPB_BADTOKEN);
	
	if ((lapb=(lapb_cb*)dev->priv) == NULL){
		return LAPB_BADTOKEN;
	}
	
	parms->t1      = lapb->t1 / HZ;
	parms->t2      = lapb->t2 / HZ;
	parms->n2      = lapb->n2;
	parms->n2count = lapb->n2count;
	parms->state   = lapb->state;
	parms->window  = lapb->window;
	parms->mode    = lapb->mode;

	if (!timer_pending(&lapb->t1timer))
		parms->t1timer = 0;
	else
		parms->t1timer = (lapb->t1timer.expires - jiffies) / HZ;

	if (!timer_pending(&lapb->t2timer))
		parms->t2timer = 0;
	else
		parms->t2timer = (lapb->t2timer.expires - jiffies) / HZ;

	return LAPB_OK;
}

int lapb_setparms(struct net_device *dev, struct lapb_parms_struct *parms)
{
	lapb_cb *lapb;

	WAN_ASSERT2((!dev),LAPB_BADTOKEN);
	WAN_ASSERT2((!parms),LAPB_BADTOKEN);
	
	
	if ((lapb=(lapb_cb*)dev->priv) == NULL)
		return LAPB_BADTOKEN;

	if (parms->t1 < 1){
		printk(KERN_INFO "%s: Error: Lapb T1 timer is invalid \n",
				dev->name);
		return LAPB_INVALUE;
	}
	
	/* If timer is zero the RR response will
	 * be instantaneous */
	if (parms->t2 < 0){
		printk(KERN_INFO "%s: Error: Lapb T2 timer is invalid \n",
				dev->name);
		return LAPB_INVALUE;
	}
	
	if (parms->n2 < 1){
		printk(KERN_INFO "%s: Error: Lapb N2 counter is invalid \n",
				dev->name);
		return LAPB_INVALUE;
	}

	lapb->t4_orig = parms->t4;
	lapb->t3_orig = parms->t3;

	if (parms->t4 && parms->t4 <= parms->t1){
		printk(KERN_INFO "%s: Error: Lapb T4 timer must be greater than T1!\n",
				dev->name);
		return LAPB_INVALUE;
	}
	
	if (parms->t4 < 1 && parms->t3 > 0){
		/* Disabled T4 timer but enabled T3 timer 
		 * default T4 to T3 and set T3 to 1, thus
	 	* if line is inactive T3 time, disconnect */
		parms->t4 = parms->t3;
		parms->t3 = 1;
		
	}else if (parms->t4 == 0 && parms->t3 == 0){
		/* Disabled both T4 and T3 Do nothing */
		lapb->t4_orig=0;
	
	}else if (parms->t3 > parms->t4){
		/* Both T4 and T3 are enabled,
		 * make T3 multiple of T4 */
		parms->t3 = (unsigned int)(parms->t3/parms->t4);

	}else if (parms->t3 == 0 && parms->t4 > 0){
		/* T3 timer is disabled, and t4
		 * timer is enabled */
		lapb->t3_orig=0;
		
	}else{
		printk(KERN_INFO "%s: Error: Lapb T3 and T4 values invalid! T3 must be greater than T4\n",
				dev->name);
		return LAPB_INVALUE;
	}

		
	if (lapb->state == LAPB_STATE_DISCONNECTED) {
		if (parms->mode & LAPB_EXTENDED) {
			if (parms->window < 1 || parms->window > 127)
				return LAPB_INVALUE;
		} else {
			if (parms->window < 1 || parms->window > 7)
				return LAPB_INVALUE;
		}

		lapb->mode    = parms->mode;
		lapb->window  = parms->window;
	}

	lapb->t1    = parms->t1 * HZ;
	lapb->t2    = parms->t2 * HZ;
	lapb->n2    = parms->n2;
	lapb->t4    = parms->t4 * HZ;

	/* T3 is not a timer really, we are simulating it
	 * it is really a counter, but to the user its
	 * a timer. Thus do not multiply it with HZ */
	lapb->t3    = parms->t3;
	dev->mtu    = parms->mtu;
	lapb->mtu   = parms->mtu;

	printk(KERN_INFO "%s: Lapb T1=%i T2=%i T3=%i T4=%i\n",
			dev->name,lapb->t1,lapb->t2,lapb->t3,lapb->t4);

	strncpy(lapb->label, parms->label, WAN_IF_LABEL_SZ);

	/* X25_SW */
	strncpy(lapb->virtual_addr, parms->virtual_addr, WAN_ADDRESS_SZ);
	strncpy(lapb->real_addr, parms->real_addr, WAN_ADDRESS_SZ);
		
	return LAPB_OK;
}


static int lapb_if_send (struct sk_buff* skb, struct net_device* dev)
{
	lapb_cb *lapb = dev_to_lapb(dev);
	int err=0;

	netif_stop_queue(dev);
	
	if (lapb->state != LAPB_STATE_CONNECTED){ 
			
		printk(KERN_INFO "Lapb: Invalid LABP State!\n");

		if (lapb->x25_link){
			err = -EINVAL;
		}else{
			netif_start_queue(dev);
			KFREE_SKB(skb);	
			err = 0;
		}
		goto lapb_kick_exit;
	}

	if (skb_tailroom(skb) < 4 || skb_headroom(skb) < 6){
		printk(KERN_INFO "Lapb: No skb room for header H %i T %i!!!\n",
				skb_headroom(skb), skb_tailroom(skb));
		netif_start_queue(dev);
		
		if (lapb->x25_link){
			err = -EINVAL;
		}else{
			LAPB_SKB_INC(skb->truesize);
			KFREE_SKB(skb);
			err = 0;
		}
		goto lapb_kick_exit;
	}
	
	if (skb_queue_len(&lapb->write_queue) >= lapb->window){
		
		if (lapb->x25_link){
			//printk(KERN_INFO "Lapb: Queue Full!!\n");
			err = -EINVAL;
		}else{
			//printk(KERN_INFO "Lapb: TCPIP - Queue Full!!\n");
			err=1;
		}
		goto lapb_kick_exit;
	}

#ifdef __DEBUG_DROP_X25_PACKET_
	if (skb->len >= 3){
		unsigned short lcn=((skb->data[0]&0x0F)<<8) | skb->data[1];
		if (skb->data[2] == 0x13 && lcn == 10){
			printk(KERN_INFO "%s: Dropping Clear Call :)\n",dev->name);
			LAPB_SKB_INC(skb->truesize);
			KFREE_SKB(skb);
			err=0;
			netif_start_queue(dev);
			goto lapb_kick_exit;
		}
	}	
#endif

	wan_skb_unlink(skb);
	
	LAPB_SKB_INC(skb->truesize);
	
	skb_queue_tail(&lapb->write_queue, skb);
	netif_start_queue(dev);
	dev->trans_start = jiffies;

lapb_kick_exit:
	lapb_trigger_kick(lapb,1);
	return err;
}

//Called by the stack via prot_hook in lapb_cb
static int lapb_data_received(struct net_device *dev, struct sk_buff *skb)
{
	lapb_cb *lapb = dev_to_lapb(dev);
	
	LAPB_SKB_INC(skb->truesize);
	
	if (lapb == NULL){
		KFREE_SKB(skb);
		return LAPB_OK;
	}

	if (test_bit(LAPB_UNREG_CRIT, &lapb->critical)){
		KFREE_SKB(skb);
		return LAPB_OK;
	}

	lapb_data_input(lapb, skb);

	return LAPB_OK;
}

/*======================================================
 * lapb_connect_confirmation
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is connected and that it should
 * start its connection process.
 */
void lapb_connect_confirmation(lapb_cb *lapb, int reason)
{
	if (netif_queue_stopped(lapb->dev)){
		netif_wake_queue(lapb->dev);
	}

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_lapb_connect_confirm))
			x25_protocol.x25_lapb_connect_confirm(lapb->x25_link,reason);
	}
}
/*======================================================
 * lapb_connect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is connected and that it should
 * start its connection process.
 */
void lapb_connect_indication(lapb_cb *lapb, int reason)
{

	if (netif_queue_stopped(lapb->dev)){
		netif_wake_queue(lapb->dev);
	}

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_lapb_connect_indication))
			x25_protocol.x25_lapb_connect_indication(lapb->x25_link,reason);
	}
}

/*======================================================
 * lapb_disconnect_confirmation
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is disconnected and that it should
 * drop the connection.
 */
void lapb_disconnect_confirmation(lapb_cb *lapb, int reason)
{

	netif_stop_queue(lapb->dev);

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_lapb_disconnect_confirm))
			x25_protocol.x25_lapb_disconnect_confirm(lapb->x25_link,reason);
	}
}

/*======================================================
 * lapb_disconnect_indication
 *
 * Indicate to the upper layer (x25) that the
 * lapb is layer is disconnected and that it should
 * drop the connection.
 */
void lapb_disconnect_indication(lapb_cb *lapb, int reason)
{

	netif_stop_queue(lapb->dev);

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_lapb_disconnect_indication))
			x25_protocol.x25_lapb_disconnect_indication(lapb->x25_link,reason);
	}
}

/*======================================================
 * lapb_data_indication
 *
 * Pass any lapb information packets up to the
 * higher layer (X25 or IP stack).
 *
 */

int lapb_data_indication(lapb_cb *lapb, struct sk_buff *skb)
{
	int err=0;
	int len;
	
	len=skb->truesize;

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_receive_pkt)){
			err = x25_protocol.x25_receive_pkt(lapb->x25_link,skb);
			if (!err){
				LAPB_SKB_DEC(len);
			}
			return err;
		}else{
			printk(KERN_INFO "%s: Failed to send packet to X25, protocol not registred!\n",
					lapb->dev->name);
			KFREE_SKB(skb);
		}
	}else{
		LAPB_SKB_DEC(len);
		skb->dev = lapb->dev;
		skb->protocol = htons(ETH_P_IP);
		netif_rx(skb);
	}
	return 0;
}

/*======================================================
 * lapb_data_transmit
 *
 * Transmitt the lapb packet to the lower layer,
 * vi its master (tx_dev) device.
 * 
 */
int lapb_data_transmit(lapb_cb *lapb, struct sk_buff *skb)
{
	int err;
	int len;
	
	len=skb->truesize;
	
	wan_skb_unlink(skb);
	skb->dev = lapb->tx_dev;
	skb->protocol = htons(ETH_P_X25);

	if (!(lapb->tx_dev->flags & IFF_UP)){
		printk(KERN_INFO "%s: Error: Lapb tx device %s Down!: data lost\n",
				lapb->dev->name, lapb->tx_dev->name);
		KFREE_SKB(skb);
		return 0;
	}

	spin_lock(&lapb->tx_lock);
	
	if (netif_queue_stopped(lapb->tx_dev)){
#if LAPB_DEBUG
		printk(KERN_INFO "%s: Tx busy, tx device stopped %s !\n",
				lapb->dev->name,lapb->tx_dev->name);
#endif

		spin_unlock(&lapb->tx_lock);
		return 1;
	}
	
	
	err = lapb->tx_dev->hard_start_xmit(skb, lapb->tx_dev);
#if LAPB_DEBUG
	if (err != 0){
		printk(KERN_INFO "%s: Tx error, failed to xmit on %s\n",
				lapb->dev->name,lapb->tx_dev->name);
	}
#endif

	if (!err){
		LAPB_SKB_DEC(len);
	}

	spin_unlock(&lapb->tx_lock);

	return err;
}


void lapb_mark_bh (struct net_device *dev)
{
	lapb_cb *lapb = dev->priv;

	if (!lapb)
		return;

	if (test_bit (LAPB_UNREG_CRIT, &lapb->critical))
		return;
	
	if (test_and_clear_bit(LAPB_AWAITING_KICK,&lapb->critical)){
#if LAPB_DEBUG
		printk(KERN_INFO "%s: LAPB BH Kick from lower layer !\n",
				lapb->dev->name);
#endif

		tasklet_schedule(&lapb->lapb_task);
	}
}


int register_lapb_x25_protocol (struct lapb_x25_register_struct * x25_reg)
{
	memcpy(&x25_protocol,x25_reg,sizeof(struct lapb_x25_register_struct));
	REG_PROTOCOL_FUNC(x25_protocol);
	return 0;
}

void unregister_lapb_x25_protocol (void)
{
	UNREG_PROTOCOL_FUNC(x25_protocol);
}

void lapb_kickstart_tx (struct net_device *dev)
{
	lapb_cb *lapb;
	struct sk_buff *newskb=NULL;
	
	WAN_ASSERT1((dev==NULL));
	
	lapb = dev->priv;
	
	if (!lapb)
		return;

	if (lapb->condition & LAPB_RNR_SENT_CONDITION && skb_queue_len(&lapb->rx_queue)){
	
		while ((newskb=skb_dequeue(&lapb->rx_queue)) != NULL){
			if (lapb_data_indication(lapb, newskb) != 0){
				skb_queue_head(&lapb->rx_queue,newskb);
				break;
			}
		}

		if (skb_queue_len(&lapb->rx_queue) < LAPB_RX_Q_L_LIMIT){
			lapb->condition &= ~LAPB_RNR_SENT_CONDITION;
#if LAPB_DEBUG
			printk(KERN_INFO "%s: Lapb kickstart sending RR!\n",dev->name);
#endif
			lapb_send_control(lapb, LAPB_RR, LAPB_POLLOFF, LAPB_RESPONSE);
		}
	}

	return;
}

void lapb_set_mtu (struct net_device *dev, unsigned int mtu)
{
	lapb_cb *lapb;;

	WAN_ASSERT1((dev==NULL));
	lapb = dev->priv;
	
	
	if (!lapb)
		return;

	dev->mtu = mtu;
	lapb->mtu = mtu;
	return;
}

void lapb_link_up (struct net_device *dev)
{
	lapb_cb *lapb;
	
	if (!dev)
		return;

	if (!(lapb=dev->priv))
		return;

	if (test_bit(LAPB_DOWN,&lapb->critical)){
		return;
	}

	lapb->condition = 0x00;
	lapb->n2count   = 0;
	lapb_clear_queues(lapb);

	lapb_stop_t1timer(lapb);
	lapb_start_t1timer(lapb);
	lapb_stop_t2timer(lapb);

	/* Do not actually try to establish connection 
	 * here.  Since frame relay needs time to start up
	 * let the T1 timer expire and retry the 
	 * startup sequence */
	lapb_state_change(lapb,LAPB_STATE_CONNECTING);
	return;
}

void lapb_link_down (struct net_device *dev)
{
	lapb_cb *lapb;
	
	if (!dev)
		return;

	if (!(lapb=dev->priv))
		return;

	lapb_state_change(lapb,LAPB_STATE_DISCONNECTED);
	lapb_stop_t1timer(lapb);
	lapb_stop_t2timer(lapb);
	lapb_stop_t4timer(lapb);

	lapb_disconnect_indication(lapb, 0);
	return;
}

void lapb_link_unregister (struct net_device *dev)
{
	lapb_cb *lapb;
	unsigned long flags;
	
	WAN_ASSERT1((dev==NULL));
	lapb = dev->priv;
	
	if (lapb){
		spin_lock_irqsave(&lapb_spin_lock,flags);
		lapb->x25_link=NULL;
		spin_unlock_irqrestore(&lapb_spin_lock,flags);
	}
}

void lapb_get_active_inactive(struct net_device *dev, wp_stack_stats_t *wp_stats)
{
	lapb_cb *lapb;

	WAN_ASSERT1((!dev));
	WAN_ASSERT1((!wp_stats));
	
	lapb = dev->priv;

	if (!lapb)
		return;

	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_get_active_inactive)){
			x25_protocol.x25_get_active_inactive(lapb->x25_link,wp_stats);
		}
	}

	if (lapb->state == LAPB_STATE_CONNECTED){
		wp_stats->lapb_active++;
	}else{
		wp_stats->lapb_inactive++;
	}

	return;
}


int lapb_get_map(struct net_device *dev, struct seq_file* m)
{
	lapb_cb*	lapb = dev->priv;
	char		line[200];

	if (test_bit(LAPB_UNREG_CRIT, &lapb->critical)){
		return 0;
	}

	PROC_ADD_LINE(m,":%s:%c", 
		 dev->name,
		 test_bit(LAPB_DOWN,&lapb->critical) ? 'd' : 
		 (lapb->state == LAPB_STATE_CONNECTED) ? '*':' ');
	
	if (lapb->x25_link){
		if (IS_FUNC_CALL(x25_protocol,x25_get_map)){
			x25_protocol.x25_get_map(lapb->x25_link,m);
			return 0;
		}
	}

	PROC_ADD_LINE(m,"%15s%s\n", lapb->label, line);

	return 0;
}

static int __init lapb_init(void)
{
	struct wanpipe_lapb_register_struct wanpipe_lapb_req;
	int err=0;

	lapb_list=NULL;

	atomic_set(&lapb_skb_alloc,0);
	atomic_set(&lapb_mem_alloc,0);
	UNREG_PROTOCOL_FUNC(x25_protocol);

	wanpipe_lapb_req.lapb_register = lapb_register;
	wanpipe_lapb_req.lapb_unregister = lapb_unregister;
	wanpipe_lapb_req.lapb_getparms = lapb_getparms;
	wanpipe_lapb_req.lapb_setparms = lapb_setparms;
	wanpipe_lapb_req.lapb_mark_bh = lapb_mark_bh;
	wanpipe_lapb_req.lapb_x25_register = lapb_x25_register;
	wanpipe_lapb_req.lapb_x25_unregister = lapb_x25_unregister;
	wanpipe_lapb_req.lapb_x25_setparms = lapb_x25_setparms;
	wanpipe_lapb_req.lapb_link_up = lapb_link_up;
	wanpipe_lapb_req.lapb_link_down = lapb_link_down;
	wanpipe_lapb_req.lapb_rx = lapb_data_received;
	wanpipe_lapb_req.lapb_get_map = lapb_get_map;
	wanpipe_lapb_req.lapb_get_active_inactive=lapb_get_active_inactive;

	spin_lock_init(&lapb_spin_lock);
	
	if (WANPIPE_VERSION_BETA){
		DEBUG_EVENT("WANPIPE LAPB for Linux. Beta %s.%s %s %s\n",
			WANPIPE_VERSION, WANPIPE_SUB_VERSION, 
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
	}else{
		DEBUG_EVENT("WANPIPE LAPB for Linux. Stable %s.%s %s %s\n",
			WANPIPE_VERSION, WANPIPE_SUB_VERSION, 
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
	}          

	err = register_wanpipe_lapb_protocol(&wanpipe_lapb_req);
	if (err != 0)
		return err;

#ifdef CONFIG_PROC_FS
	err = lapb_proc_init();
#endif
	return err;
}

EXPORT_SYMBOL(register_lapb_x25_protocol);
EXPORT_SYMBOL(unregister_lapb_x25_protocol);
EXPORT_SYMBOL(lapb_kickstart_tx);
EXPORT_SYMBOL(lapb_set_mtu);
EXPORT_SYMBOL(lapb_link_unregister);

MODULE_AUTHOR("Nenad Corbic <ncorbic@sangoma.com>");
MODULE_DESCRIPTION("The LAPB Link Access Procedure B link layer protocol");
MODULE_LICENSE("GPL");

void __exit lapb_exit (void)
{

#ifdef CONFIG_PROC_FS
	 lapb_proc_cleanup();
#endif
	 unregister_wanpipe_lapb_protocol();

	 printk(KERN_INFO "WANPIPE LAPB: Unloading module\n");

	 if (atomic_read(&lapb_mem_alloc)){
		printk(KERN_INFO "X25: Warning releasing module with used memory=%i\n",
				atomic_read(&lapb_mem_alloc));
	 }
	 if (atomic_read(&lapb_skb_alloc)){
		printk(KERN_INFO "X25: Warning releasing module with used skb memory=%i\n",
				atomic_read(&lapb_skb_alloc));
	 }
}

module_init(lapb_init);
module_exit(lapb_exit);

