/*
 *	LAPB release 002
 *
 *	This code REQUIRES 2.1.15 or higher/ NET3.038
 *
 *	This module:
 *		This module is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_lapb.h"

extern struct lapb_x25_register_struct x25_protocol;

/* 
 *  This procedure is passed a buffer descriptor for an iframe. It builds
 *  the rest of the control part of the frame and then writes it out.
 */
static int lapb_send_iframe(lapb_cb *lapb, struct sk_buff *skb, int poll_bit)
{
	unsigned char *frame;

	if (skb == NULL)
		return -EINVAL;

	if (skb_tailroom(skb) < 4){
		KFREE_SKB(skb);
		printk(KERN_INFO "BAD ROOM IN SEND FRAME\n");
		return -EINVAL;
	}

	if (lapb->mode & LAPB_EXTENDED) {
		frame = skb_push(skb, 2);

		frame[0] = LAPB_I;
		frame[0] |= lapb->vs << 1;
		frame[1] = (poll_bit) ? LAPB_EPF : 0;
		frame[1] |= lapb->vr << 1;
	} else {
		frame = skb_push(skb, 1);

		*frame = LAPB_I;
		*frame |= (poll_bit) ? LAPB_SPF : 0;
		*frame |= lapb->vr << 5;
		*frame |= lapb->vs << 1;
	}

#if LAPB_DEBUG > 1
	printk(KERN_DEBUG "lapb: (%p) S%d TX I(%d) S%d R%d\n", lapb->dev, lapb->state, poll_bit, lapb->vs, lapb->vr);
#endif

	return lapb_transmit_buffer(lapb, skb, LAPB_COMMAND,0);	
}

void lapb_kick(unsigned long lapb_data)
{
#define lapb ((lapb_cb *)lapb_data)
	struct sk_buff *skb, *skbn;
	volatile unsigned short modulus, start, end;
	int err=0;

	
	if (test_and_set_bit(LAPB_SMP_CRIT,&lapb->critical)){
		printk(KERN_INFO "CRITICAL IN LAPB KICK !!!---!!!\n");
		return;
	}

	if (test_bit(LAPB_UNREG_CRIT, &lapb->critical)){
		clear_bit(LAPB_AWAITING_KICK,&lapb->critical);
		clear_bit(LAPB_FRAME_SENT,&lapb->critical);
		clear_bit(LAPB_BH_RUNNING,&lapb->critical);
		clear_bit(LAPB_SMP_CRIT,&lapb->critical);
		return;
	}

	spin_lock(&lapb->bh_lock);
	
	++lapb->kick_cnt;
	
	clear_bit(LAPB_FRAME_SENT,&lapb->critical);
	
lapb_kick_send_async:
	while ((skb=skb_dequeue(&lapb->async_queue)) != NULL){

		/* Fix to include all packets that contain
		 * VR numbers */
		switch ((skb->data[1]&0x0F)){
			case LAPB_RR:
			case LAPB_RNR:
			case LAPB_REJ:
				skb->data[1] &= 0x1F;
				skb->data[1] |= (lapb->vr << 5);
				break;
		}
		err = lapb_data_transmit(lapb, skb);

		if (err != 0){
			//printk(KERN_INFO "%s:Failed to send lapb Async Message!\n",
			//		lapb->dev->name);
			if (lapb->state == LAPB_STATE_CONNECTED){
				skb_queue_head(&lapb->async_queue,skb);
				set_bit(LAPB_AWAITING_KICK,&lapb->critical);
				goto lapb_kick_exit;
			}else{
				/* NC: Jul 05 2005: Memory Leak Bug fix.
                                 * During connecting state, the hardware
                                 * below might be in disconnected state,
                                 * thus it will not issue a kick callback.
                                 *
                                 * Therefore, in case of failed tx attempt don't
                                 * free the Async packet and with for the timer
                                 * to send another one */
				KFREE_SKB(skb);
			}
		}
	}

	if (lapb->state != LAPB_STATE_CONNECTED)
		goto lapb_kick_exit;

	if (test_bit(LAPB_REQUEUING,&lapb->critical)){
		printk(KERN_INFO "(Debug): Lapb requeuing skip kick!\n");
		goto lapb_kick_exit;
	}
	
	modulus = (lapb->mode & LAPB_EXTENDED) ? LAPB_EMODULUS : LAPB_SMODULUS;

//	WARNING: This line doesn't work in SMP systems, because there 
//	         can be a racing condition when the queue is empty and
//	         
//	start = (skb_peek(&lapb->ack_queue) == NULL) ? lapb->va : lapb->vs;

        start = lapb->vs;
	end   = (lapb->va + lapb->window) % modulus;
	
	if (!(lapb->condition & LAPB_PEER_RX_BUSY_CONDITION) &&
	    start != end                                &&
	    skb_peek(&lapb->write_queue) != NULL){

		/*
		 * Dequeue the frame and copy it.
		 */
		skb  = skb_dequeue(&lapb->write_queue);

		do {
			if (skb_queue_len(&lapb->async_queue)){
				//printk(KERN_INFO "Lapb kick re-sending async!\n");
				skb_queue_head(&lapb->write_queue, skb);
				goto lapb_kick_send_async;
			}

			
			if ((skbn = skb_clone(skb, GFP_ATOMIC)) == NULL) {
				skb_queue_head(&lapb->write_queue, skb);
				break;
			}
			LAPB_SKB_INC(skbn->truesize);


			if (skb->sk != NULL)
				skb_set_owner_w(skbn, skb->sk);

			/*
			 * Transmit the frame copy.
			 */
			err = lapb_send_iframe(lapb, skbn, LAPB_POLLOFF);
			
			if (err == 0){

				//printk(KERN_DEBUG "LAPB KICK PACKET OK: Vs=%i\n",
				//		lapb->vs);

				lapb_update_stats (lapb, LAPB_I, skb->len, 1);
	
				lapb->vs = (lapb->vs + 1) % modulus;
				/* Requeue the original data frame */
				skb_queue_tail(&lapb->ack_queue, skb);
	
				set_bit(LAPB_FRAME_SENT,&lapb->critical);

			}else if (err == 1){
				/* Failed to Tx data, thus re-queue frame */
				DEBUG_RX("Lapb: Failed to TX Data to Lower layer! Vs=%i\n",
						lapb->vs);
				skb_queue_head(&lapb->write_queue, skb);

				KFREE_SKB(skbn);
				
				set_bit(LAPB_AWAITING_KICK,&lapb->critical);
				break;
			}else{
				/* Major Error */ 
				printk(KERN_INFO "Lapb: Failed to TX Data Major Error! Vs=%i\n",
						lapb->vs);
				skb_queue_head(&lapb->write_queue, skb);
				
				KFREE_SKB(skbn);

				set_bit(LAPB_AWAITING_KICK,&lapb->critical);
				break;
			}
			
		} while (lapb->vs != end && (skb = skb_dequeue(&lapb->write_queue)) != NULL);

		/* Only stop the RR response if we successfully sent
		 * a packet. Otherwise let the RR clear the window */
		if (test_bit(LAPB_FRAME_SENT,&lapb->critical)){

			lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;

			if (!lapb_t1timer_running(lapb))
				lapb_start_t1timer(lapb);
			
			clear_bit(LAPB_FRAME_SENT,&lapb->critical);


			if (netif_queue_stopped(lapb->dev)){
				netif_wake_queue(lapb->dev);
			}
			//The Window is now open, kick the upper
			//layers so that they can send more data.
			if (lapb->x25_link){
				if (IS_FUNC_CALL(x25_protocol,x25_lapb_bh_kick))
					x25_protocol.x25_lapb_bh_kick(lapb->x25_link);
			}
		}
	}

lapb_kick_exit:
	
	spin_unlock(&lapb->bh_lock);
	clear_bit(LAPB_SMP_CRIT,&lapb->critical);
	clear_bit(LAPB_BH_RUNNING,&lapb->critical);
#undef lapb
}

int lapb_transmit_buffer(lapb_cb *lapb, struct sk_buff *skb, int type, char async)
{
	unsigned char *ptr;
	//int err;

	ptr = skb_push(skb, 1);

	if (lapb->mode & LAPB_MLP) {
		if (lapb->mode & LAPB_DCE) {
			if (type == LAPB_COMMAND)
				*ptr = LAPB_ADDR_C;
			if (type == LAPB_RESPONSE)
				*ptr = LAPB_ADDR_D;
		} else {
			if (type == LAPB_COMMAND)
				*ptr = LAPB_ADDR_D;
			if (type == LAPB_RESPONSE)
				*ptr = LAPB_ADDR_C;
		}
	} else {
		if (lapb->mode & LAPB_DCE) {
			if (type == LAPB_COMMAND)
				*ptr = LAPB_ADDR_A;
			if (type == LAPB_RESPONSE)
				*ptr = LAPB_ADDR_B;
		} else {
			if (type == LAPB_COMMAND)
				*ptr = LAPB_ADDR_B;
			if (type == LAPB_RESPONSE)
				*ptr = LAPB_ADDR_A;
		}
	}

#if LAPB_DEBUG > 2
	printk(KERN_DEBUG "lapb: (%p) S%d TX %02X %02X %02X\n", lapb->dev, lapb->state, skb->data[0], skb->data[1], skb->data[2]);
#endif

#if 1
	if (async){
		skb_queue_tail(&lapb->async_queue,skb);
		lapb_trigger_kick(lapb,1);
		return 0;
	}

	return lapb_data_transmit(lapb, skb);
#else

	err = lapb_data_transmit(lapb, skb);
	
	if (err && async){
		skb_queue_tail(&lapb->async_queue,skb);
		lapb_trigger_kick(lapb,1);
	}
	return err;
#endif
}

void lapb_establish_data_link(lapb_cb *lapb)
{
	lapb->condition = 0x00;
	lapb->n2count   = 0;

	/* Nenad Corbic: Initailize all queues before
	 * starting the link */
	lapb_clear_queues(lapb);
	
	if (lapb->mode & LAPB_EXTENDED) {
#if LAPB_DEBUG > 1
		printk(KERN_DEBUG "lapb: (%p) S%d TX SABME(1)\n", lapb->dev, lapb->state);
#endif
		lapb_send_control(lapb, LAPB_SABME, LAPB_POLLON, LAPB_COMMAND);
	} else {
#if LAPB_DEBUG > 1
		printk(KERN_DEBUG "lapb: (%p) S%d TX SABM(1)\n", lapb->dev, lapb->state);
#endif
		lapb_send_control(lapb, LAPB_SABM, LAPB_POLLON, LAPB_COMMAND);
	}

	lapb_start_t1timer(lapb);
	lapb_stop_t2timer(lapb);
}

void lapb_enquiry_response(lapb_cb *lapb)
{
#if LAPB_DEBUG > 1
	printk(KERN_DEBUG "lapb: (%p) S%d TX RR(1) R%d\n", lapb->dev, lapb->state, lapb->vr);
#endif
	lapb_send_control(lapb, LAPB_RR, LAPB_POLLON, LAPB_RESPONSE);

	lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
}

void lapb_timeout_response(lapb_cb *lapb)
{
#if LAPB_DEBUG > 1
	printk(KERN_DEBUG "lapb: (%p) S%d TX RR(0) R%d\n", lapb->dev, lapb->state, lapb->vr);
#endif

	if (skb_queue_len(&lapb->rx_queue) > LAPB_RX_Q_H_LIMIT){
		lapb_send_control(lapb, LAPB_RNR, LAPB_POLLOFF, LAPB_RESPONSE);
	}else{
		lapb_send_control(lapb, LAPB_RR, LAPB_POLLOFF, LAPB_RESPONSE);
	}

	lapb->condition &= ~LAPB_ACK_PENDING_CONDITION;
}

void lapb_check_iframes_acked(lapb_cb *lapb, unsigned short nr)
{
	if (lapb->vs == nr) {
		lapb_frames_acked(lapb, nr);
		lapb_stop_t1timer(lapb);
		lapb->n2count = 0;
	} else {
		if (lapb->va != nr) {
			lapb_frames_acked(lapb, nr);
			lapb_start_t1timer(lapb);
		}
	}
}

void lapb_check_need_response(lapb_cb *lapb, int type, int pf)
{
	if (type == LAPB_COMMAND && pf){
		lapb_enquiry_response(lapb);
	}
}

