/*****************************************************************************
* x25_iface.c	WANPIPE(tm) X25 ANNEXG
*
* Authors: 	Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 2001-2003 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 13, 2003 	Nenad Corbic	Added SIOC_X25_SET_LCN_LABEL ioctl
* May 03, 2001  Nenad Corbic	Initial Release
*/
#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"

#define MAX_LINK_RX_Q_LEN 10
#define MAX_TAIL_ROOM 16
#define MAX_X25_LINKS 2000
static int x25_if_send (struct sk_buff*, struct net_device*);
static int x25_if_init(struct net_device *dev);

atomic_t x25_skb_alloc;
atomic_t x25_mem_alloc;

x25_link_t *x25_link_list;
rwlock_t x25_link_lock;

struct x25_dsp_register_struct x25_dsp_protocol;

static unsigned char x25_link_num[MAX_X25_LINKS];

/*
 *	Free an allocated lapb control block. This is done to centralise
 *	the MOD count code.
 */
static void x25_free_dev(struct net_device *dev)
{
	x25_lcn_t *x25_lcn=dev_to_x25(dev);
	
	if (x25_lcn){
		unsigned long timeout=jiffies;
		
		dev->priv=NULL;
		X25_PUT(x25_lcn);
		
x25_free_dev_wait:
		if (atomic_read(&x25_lcn->refcnt)){
			printk(KERN_INFO "%s: Error x25 lcn is still in use %i, waiting 2 sec...\n",
					dev->name,atomic_read(&x25_lcn->refcnt));
		
			if ((jiffies-timeout)<(HZ*2)){
				schedule();
				goto x25_free_dev_wait;
			}

			printk(KERN_INFO "%s: Warning: proceeding with deallocation: refcnt=%i\n",
					dev->name,atomic_read(&x25_lcn->refcnt));
		}

		/* NOTE: We do not try to do X25_PUT() on dev, since
		 * it has already been removed, and we are sure that
		 * its refcnt is now 0, thus it is safe to just clear
		 * the dev pointer */
		if (x25_lcn->dev){
			clear_bit(LCN_DEV,&x25_lcn->used);
			x25_lcn->dev=NULL;
		}

		if (x25_lcn->x25_link){
			clear_bit(LCN_X25_LINK,&x25_lcn->used);
			X25_PUT(x25_lcn->x25_link);
			x25_lcn->x25_link=NULL;
		}

		if (x25_lcn->used){
			printk(KERN_INFO "%s: Major error x25 lcn used resources= 0x%lx\n",
					dev->name,x25_lcn->used);
		}
		KFREE(x25_lcn);
	}


	if (atomic_read(&dev->refcnt)){
		printk(KERN_INFO "%s: Major error dev is still in use %i\n",
					dev->name,atomic_read(&dev->refcnt));
	}

	wan_netif_free(dev);
}

/* X25_SW
 * Search through all X.25 links and find the X.25 link that configured 
 * for the same virtual address.
 *
 * NC. Fri Oct 11 18:04:25 EDT 2002 
 *     Changed as per Steves request to do a string
 *     search for the virtual address in incoming
 *     called address.
 *     
 */
x25_link_t* x25_get_switch_link(unsigned char* addr)
{
	x25_link_t*	x25_link = NULL;
	unsigned long   flags;
	
	WP_READ_LOCK(&x25_link_lock,flags);
	
	for (x25_link = x25_link_list; x25_link;  
			x25_link = x25_link->next){

#if 1
		if (!strlen(x25_link->virtual_addr)){
			continue;
		}
		
		if (strlen(x25_link->virtual_addr) > strlen(addr)){
			continue;
		}
		
		if (strstr(addr,x25_link->virtual_addr) != NULL){
			WP_READ_UNLOCK(&x25_link_lock,flags);
			return x25_link;
		}
#else		
		if (strcmp(addr, x25_link->virtual_addr) == 0){
			WP_READ_UNLOCK(&x25_link_lock,flags);
			return x25_link;
		}
#endif
	}
	
	WP_READ_UNLOCK(&x25_link_lock,flags);
	
	return NULL;
}

static void x25_free_link(x25_link_t *x25_link)
{
	unsigned long timeout=jiffies;
	int i;

	if (x25_link->link_num != -1){
		clear_bit(0,(unsigned long*)&x25_link_num[x25_link->link_num]);
		x25_link->link_num=-1;
	}

	X25_PUT(x25_link);

x25_free_link_wait:
	if (atomic_read(&x25_link->refcnt)){
		printk(KERN_INFO "%s: Error x25 link use count %i, waiting 2 sec...\n",
				x25_link->name, atomic_read(&x25_link->refcnt));
		
		if ((jiffies-timeout) < (HZ*2)){
			schedule();
			goto x25_free_link_wait;
		}

		printk(KERN_INFO "%s: Warning continuing with deallocation: refcnt=%i\n",
				x25_link->name, atomic_read(&x25_link->refcnt));
		atomic_set(&x25_link->refcnt,1);
	}

	for (i=0;i<MAX_PROC_EVENTS;i++){
		if (x25_link->proc_event_log[i]){
			kfree(x25_link->proc_event_log[i]);
			X25_MEM_DEC(MAX_PROC_EVENT_SIZE);
			x25_link->proc_event_log[i]=NULL;
		}
	}

	KFREE(x25_link);

#ifndef LINUX_2_6
	MOD_DEC_USE_COUNT;
#endif
}

static void x25_remove_link(x25_link_t *x25_link)
{
	x25_link_t *cur_link;
	x25_link_t *prev_link;
	unsigned long flags;
	
	WP_READ_LOCK(&x25_link_lock,flags);
	
	if ((cur_link = x25_link_list) == x25_link) {
		x25_link_list = cur_link->next;
		WP_READ_UNLOCK(&x25_link_lock,flags);
		return;
	}

	while (cur_link != NULL && cur_link->next != NULL) {
		prev_link=cur_link;
		cur_link = cur_link->next;
		if (cur_link == x25_link) {
			prev_link->next = x25_link->next;
			WP_READ_UNLOCK(&x25_link_lock,flags);
			return;
		}
	}

	WP_READ_UNLOCK(&x25_link_lock,flags);
}


/*==============================================================
 * x25_remove_dev
 * 
 * 	Remove the x25 svc network device from the x25 link
 * 	device list.
 * 
 */
static void x25_remove_dev(x25_link_t *x25_link,struct net_device *dev)
{
	struct net_device *cur_dev;
	struct net_device *prev_dev;
	unsigned long flags,timeout=jiffies;


x25_remove_dev_wait:
#ifdef LINUX_2_6
	if (atomic_read(&dev->refcnt) > 2){
#else
	if (atomic_read(&dev->refcnt) != 1){
#endif
	
		struct in_device *in_dev;
		if ((in_dev =in_dev_get(dev)) && in_dev->ifa_list){
			goto x25_remove_ok;
		}
		
		if (net_ratelimit()){
			printk(KERN_INFO "%s: Error device use count %i, waiting 2 sec...\n",
				dev->name, atomic_read(&dev->refcnt));
		}
		if ((jiffies-timeout) < (HZ*2)){
			schedule();
			goto x25_remove_dev_wait;
		}

		printk(KERN_INFO "%s: Warning continuing with deallocation: refcnt=%i\n",
				dev->name, atomic_read(&dev->refcnt));

#ifndef LINUX_2_6
		atomic_set(&dev->refcnt,1);
#endif
	}
x25_remove_ok:
	
	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	
	if ((cur_dev = x25_link->dev_list) == dev) {
		x25_link->dev_list = *(struct net_device**)cur_dev->priv;
	
		WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);

		unregister_netdev(dev);

		x25_link->dev_cnt--;
		return;
	}

	while (cur_dev != NULL && *(struct net_device**)cur_dev->priv != NULL) {
		prev_dev=cur_dev;
		cur_dev = *(struct net_device**)cur_dev->priv;
		if (cur_dev == dev) {
			*(struct net_device**)prev_dev->priv = *(struct net_device**)dev->priv;
		
			WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);

			unregister_netdev(dev);

			x25_link->dev_cnt--;
			return;
		}
	}

	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
	return;
}

/*==============================================================
 * x25_insert_link
 * 
 * 	Insert the x25 link into the x25 link list.
 * 	Support for multiple x25 links.
 * 
 */

static void x25_insert_link(x25_link_t *x25_link)
{
	unsigned long flags;
	
	WP_WRITE_LOCK(&x25_link_lock,flags);
	
	x25_link->next = x25_link_list;
	x25_link_list = x25_link;

	WP_WRITE_UNLOCK(&x25_link_lock,flags);
}

/*==============================================================
 * x25_insert_dev
 * 
 * 	Insert the x25 svc network interface dev into 
 * 	the x25 link device list.
 */

static void x25_insert_dev(x25_link_t *x25_link, struct net_device *dev)
{

	unsigned long flags;
	
	WP_WRITE_LOCK(&x25_link->dev_list_lock,flags);

	*(struct net_device **)dev->priv = x25_link->dev_list;
	x25_link->dev_list = dev;
	x25_link->dev_cnt++;

	WP_WRITE_UNLOCK(&x25_link->dev_list_lock,flags);
}

/*==============================================================
 * x25_link_ok
 *
 * 	Make sure that the link in question exists
 * 	in the main x25 link list.  
 * 	Used as a sanity check.
 */

static int x25_link_ok(x25_link_t *x25_link)
{
	x25_link_t *cur_link;
	unsigned long flags;
	
	WP_READ_LOCK(&x25_link_lock,flags);
	
	for ((cur_link = x25_link_list); cur_link != NULL;  cur_link = cur_link->next){
		if (cur_link == x25_link){
			WP_READ_UNLOCK(&x25_link_lock,flags);
			return X25_OK;
		}
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);
	return -EINVAL;
}

/*==============================================================
 * x25_bind_listen_to_link
 *
 * Description:
 *	The api socket layer uses this function to bind
 *	the listen socket to the x25 link.
 *
 *	The incoming call will be passed up the socket
 *	using the bounded socket id (sk_id). The user
 *	API application will then be able to accept or
 *	clear the call.
 *
 * 	If the sk_id is null, all incoming calls
 * 	will be cleared by the x25 layer. 
 * 	
 * Used by: 
 * 	api socket layer.
 * 
 */

int x25_bind_listen_to_link(struct net_device *master_dev, void *sk_id, unsigned short protocol)
{
	x25_link_t *x25_link;
	unsigned long flags;

	WP_READ_LOCK(&x25_link_lock,flags);	
	for ((x25_link = x25_link_list); x25_link != NULL;  x25_link = x25_link->next){
		if (x25_link->master == master_dev){
			break;
		}
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);	
	
	if (!x25_link)
		return -ENODEV;

	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		printk(KERN_INFO "%s: Failed to bind listen sock: link down\n",
				x25_link->name);
		return -ENODEV;
	}
	
	//FIXME: We should check here for other protocols 
	if (protocol == ETH_P_X25){
		if (x25_link->api_sk_id)
			return -EBUSY;

		sock_hold(sk_id);
		x25_link->api_sk_id = sk_id;
	}else{
		if (x25_link->dsp_sk_id)
			return -EBUSY;

		sock_hold(sk_id);
		x25_link->dsp_sk_id = sk_id;
	}

	return 0;	
}

/*==============================================================
 * x25_unbind_listen_to_link
 *
 * Description:
 *	The api socket layer uses this function to unbind
 *	the listen socket from the x25 link.
 *
 * 	When the listen socket is released, the sk_id
 * 	must be removed from the x25 link.  
 *
 * Used by: 
 * 	api socket layer.
 * 
 */

int x25_unbind_listen_from_link(void *sk_id,unsigned short protocol)
{
	x25_link_t *x25_link = NULL;
	unsigned long flags;

	WAN_ASSERT2((sk_id==NULL),-ENODEV);

	WP_READ_LOCK(&x25_link_lock,flags);	
	for ((x25_link = x25_link_list); x25_link != NULL;  x25_link = x25_link->next){
		if (protocol == ETH_P_X25){
			if (x25_link->api_sk_id == sk_id){
				break;
			}
		}else{
			if (x25_link->dsp_sk_id == sk_id){
				break;
			}
		}
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);	
	
	if (!x25_link)
		return -ENODEV;
		
	//FIXME: We should check here for other protocols 
	if (protocol == ETH_P_X25){
		if (x25_link->api_sk_id){
			__sock_put(x25_link->api_sk_id);
		}
		x25_link->api_sk_id = NULL;
	}else{
		if (x25_link->dsp_sk_id){
			__sock_put(x25_link->dsp_sk_id);
		}
		x25_link->dsp_sk_id = NULL;
	}

	return 0;	
}

/*==============================================================
 * x25_bind_api_to_svc
 *
 * Description:
 *	The api socket layer uses this function to bind
 *	the SVC data socket to the x25_lcn link.
 *
 *	All data received by the svc will be passed up 
 *	to the socket via bounded sk_id.
 *
 * Used by: 
 * 	api socket layer.
 * 
 */

int x25_bind_api_to_svc(struct net_device *master_dev, void *sk_id)
{
	x25_link_t *x25_link = NULL;
	x25_lcn_t *x25_lcn = NULL;
	unsigned long flags;

	WAN_ASSERT2((sk_id==NULL),-ENODEV);
	
	WP_READ_LOCK(&x25_link_lock,flags);	
	for ((x25_link = x25_link_list); x25_link != NULL;  x25_link = x25_link->next){
		if (x25_link->master == master_dev){
			break;
		}
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);	
	
	if (!x25_link){
		printk(KERN_INFO "X25 Bind: No master device %s found on any link\n",
				master_dev->name);
		return -ENODEV;
	}

	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		printk(KERN_INFO "%s: X25 Bind: Failed to bind to svc: x25 link down\n",
				x25_link->name);
		return -ENODEV;
	}
	
	x25_lcn = x25_get_free_api_dev(x25_link, 0xFFFF);
	if (!x25_lcn){
		DEBUG_TEST("X25 Bind: No free device found in %s !\n",x25_link->name);
		return -ENODEV;
	}

	sock_hold(sk_id);
	x25_lcn->sk_id=sk_id;
	set_bit(LCN_SK_ID,&x25_lcn->used);

	return x25_lcn->dev->ifindex;
}

/*==============================================================
 * get_svc_dev
 *
 * Description:
 *	The DSP layer uses this function to find available svc.
 *
 * Used by: 
 * 	DSP layer.
 * 
 */

struct net_device* x25_get_svc_dev(struct net_device *master_dev)
{
	struct net_device *dev;
	unsigned long flags;
	x25_link_t *x25_link;
	x25_lcn_t *x25_lcn;
	int	ret;

	WP_READ_LOCK(&x25_link_lock,flags);	
	for ((x25_link = x25_link_list); x25_link != NULL;  x25_link = x25_link->next){
		if (x25_link->master == master_dev){
			break;
		}
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);	
	
	if (!x25_link){
		return NULL;
	}

	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		printk(KERN_INFO "%s: X25 Failed to bind to dsp svc: x25 link down\n",
				x25_link->name);
		return NULL;
	}

	
	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	for (dev=x25_link->dev_list; dev; dev = *(struct net_device **)dev->priv){
		x25_lcn = dev_to_x25(dev);

		if (x25_lcn->usedby == API && 
		    x25_lcn->dsp_id &&
		    !x25_lcn->sk_id &&
		    !x25_lcn->x25_sw_lcn &&
		    x25_lcn->state == X25_ST_PKT_LVL_RDY_P1 && 
		    atomic_read(&x25_lcn->lcn) == 0 &&
		    !test_and_set_bit(0,&x25_lcn->used)){
   			
			if (IS_FUNC_CALL(x25_dsp_protocol, dsp_check_dev)) {
		        	ret = x25_dsp_protocol.dsp_check_dev(x25_lcn->dsp_id);
		                if (ret == 0){
					break;
				}
	 		}
			clear_bit(0,&x25_lcn->used);
		}
	}

	if (!dev){
		WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
		return NULL;
	}
	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
	return dev;
}


/*==============================================================
 * x25_get_call_info
 *
 * Description:
 *	The DSP layer uses this function to receive lcn/mtu/mru for 
 *	current device.
 *
 * Used by: 
 * 	DSP layer.
 * 
 */

int x25_get_call_info(struct net_device *master_dev, 
		      unsigned short* lcn,
		      unsigned short* mtu,
		      unsigned short* mru)
{
	x25_lcn_t *x25_lcn;

	if ((x25_lcn = dev_to_x25(master_dev)) == NULL){
		return -EINVAL;
	}

	if (lcn){
		*lcn = atomic_read(&x25_lcn->lcn);
	}
	if (mtu){
		*mtu = x25_lcn->tx_pkt_size;
				
	}
	if (mru){
		*mru = x25_lcn->rx_pkt_size;
	}
	return 0;
}

/***************************************************************
 *	     X25 REGISTRATION FUNCTIONS 
 ***************************************************************/


/*==============================================================
 * x25_dev_in_link
 *
 * Description:
 *	Check that the svc network interface device 
 *	exists in the x25 link device list.
 *
 * Purpose:
 *	Prevents duplicate registration.
 * 
 * Used by: 
 *	x25_register 
 */

static struct net_device *x25_dev_in_link(x25_link_t *x25_link, char *dev_name)
{
	struct net_device *tmp_dev;
	unsigned long flags;
	
	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	for (tmp_dev = x25_link->dev_list; tmp_dev != NULL; tmp_dev= *(struct net_device **)tmp_dev->priv){
		if (strcmp(tmp_dev->name,dev_name) == X25_OK){
			WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
			return tmp_dev;
		}
	}
	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
	return NULL;
}


/*==============================================================
 * x25_create_lcn
 *
 * Description:
 * 	Create and initialize the svc/lcn 
 * 	private structure.  
 *
 * 	This structure is created for each
 * 	svc on the x25 link.  It is bound into
 * 	dev->priv pointer of the svc network 
 * 	interface device.
 *
 * Used by: 
 *	x25_register 
 */

static x25_lcn_t *x25_create_lcn(void)
{
	x25_lcn_t *x25;

	KMALLOC(x25,sizeof(x25_lcn_t),GFP_KERNEL);
	if (x25 == NULL)
		return NULL;

	memset(x25, 0x00, sizeof(x25_lcn_t));

	init_timer(&x25->T11_T21_timer);
	init_timer(&x25->T12_T22_timer);
	init_timer(&x25->T13_T23_timer);
	init_timer(&x25->T16_T26_timer);

	skb_queue_head_init(&x25->write_queue);
	skb_queue_head_init(&x25->async_queue);

	spin_lock_init(&x25->lcn_lock);

	atomic_set(&x25->refcnt,0);
	
	lcn_state_change(x25,X25_ST_HDLC_NOT_IN_ABM);
	return x25;
}

void x25_bh_link_send (unsigned long data)
{
#define x25_lnk_ptr ((x25_link_t*)data)
	if (spin_is_locked(&x25_lnk_ptr->bh_lock)){
		DEBUG_TEST("%s: X25 BH Link: bh_lock locked !!!!!\n",
				x25_lnk_ptr->name);
	}
	
	spin_lock(&x25_lnk_ptr->bh_lock);
	x25_link_send(data);
	spin_unlock(&x25_lnk_ptr->bh_lock);
#undef x25_lnk_ptr
}

/*==============================================================
 * x25_create_link
 *
 * Description:
 * 	Create and initialize the x25 link 
 * 	private structure.  
 *
 * 	This structure is created for each
 * 	x25 link. Each x25 link can have multiple
 * 	svc's.  Thus it contins the svc device
 * 	link lists as well as x25 link configuration
 * 	profiles.
 *
 * Used by: 
 *	x25_register 
 */
static x25_link_t *x25_create_link(struct net_device *device)
{
	x25_link_t *x25;
	int i;
	char *buf;
	
	KMALLOC(x25,sizeof(x25_link_t),GFP_KERNEL);
	if (x25==NULL){
		WAN_MEM_ASSERT("X25 Link Alloc: ");
		return NULL;
	}

	memset(x25,0,sizeof(x25_link_t));

	for (i=0;i<MAX_PROC_EVENTS;i++){
		KMALLOC(buf,MAX_PROC_EVENT_SIZE,GFP_KERNEL);
		if (!buf){
			for (i=0;i<MAX_PROC_EVENTS;i++){
				if (x25->proc_event_log[i]){
					KFREE(x25->proc_event_log[i]);
					x25->proc_event_log[i]=NULL;
				}
			}
			KFREE(x25);
			printk(KERN_INFO "X25 Link Alloc: Failed to allocate Event Buf!\n");
			return NULL;
		}

		memset(buf,0,MAX_PROC_EVENT_SIZE);
		x25->proc_event_log[i]=buf;
	}
	
	x25->master = device;
	
	init_timer(&x25->T10_T20_timer);
	init_timer(&x25->T28_timer);

	skb_queue_head_init(&x25->async_queue);
	skb_queue_head_init(&x25->rx_queue);

	x25->map_lock=RW_LOCK_UNLOCKED;
	x25->dev_list_lock=RW_LOCK_UNLOCKED;

	spin_lock_init(&x25->bh_lock);

	atomic_set(&x25->refcnt,0);
	atomic_set(&x25->proc_event_offset,0);

#if 1
	tasklet_init(&x25->x25_task,x25_bh_link_send,(unsigned long)x25);
#else	
	x25->x25_task.sync = 0;
	x25->x25_task.routine = (void *)(void *)x25_link_send;
	x25->x25_task.data = x25;
#endif

	x25->link_num=-1;
	for (i=0;i<MAX_X25_LINKS;i++){
		if (!test_and_set_bit(0,(unsigned long*)&x25_link_num[i])){
			x25->link_num=i;
			break;
		}
	}
	
	if (x25->link_num == -1){

		for (i=0;i<MAX_PROC_EVENTS;i++){
			if (x25->proc_event_log[i]){
				KFREE(x25->proc_event_log[i]);
				x25->proc_event_log[i]=NULL;
			}
		}

		KFREE(x25);
		printk(KERN_INFO "X25: No x25 links available! Max=%i\n",
				MAX_X25_LINKS);
		return NULL;
	}	

	sprintf(x25->name,"link%04lu",x25->link_num);
	
	x25->cfg.mode = MOD8_CONFIG;
	x25->cfg.dte  = 1;

	x25->state = X25_ST_HDLC_NOT_IN_ABM;

	/* X25_SW X.25 switching */
	memset(x25->virtual_addr, '\0', WAN_ADDRESS_SZ+1);
	memset(x25->real_addr, '\0', WAN_ADDRESS_SZ+1);

#ifndef LINUX_2_6
	MOD_INC_USE_COUNT;
#endif
	X25_HOLD(x25);

	return x25;
}


/*==============================================================
 * x25_create_dev
 *
 * Description:
 * 	Allocate and initialize the svc network interface
 * 	device.  This structure will be used to bind the
 * 	svc to the TCP/IP stack, or to the API socket.
 *	
 *	The dev->priv pointer will contain the svc/lcn
 *	private structure.
 *
 * Used by: 
 *	x25_register 
 */

static struct net_device *x25_create_dev(char *dev_name)
{
	struct net_device *dev;
	int iftype = WAN_IFT_OTHER;
	int err;

	dev=wan_netif_alloc(dev_name,iftype,&err);
	if (dev == NULL){
		return NULL;
	}

	wan_atomic_set(&dev->refcnt,0);

	return dev;
}


/*==============================================================
 * x25_register
 *
 * Description:
 *	This function is called during system setup to create
 *	and initialize the x25 link and each x25 svc defined
 *	in the wanpipe configuration file.
 *
 *	For each x25 link and x25 svc the proc file
 *	entry is created which is used to option protocol
 *	and packet statistics.
 *	
 * Usedby:
 * 	Lapb layer. 
 */


int x25_register(void **x25_lnk, struct net_device *lapb_dev, char *dev_name, struct net_device **new_dev)
{
	x25_link_t *x25_link = (x25_link_t *)*x25_lnk;
	x25_lcn_t *x25_lcn;
	struct net_device *dev;
	int err;

	WAN_ASSERT2((!lapb_dev),-ENODEV);
	
	/* Create the new X25 link for this Lapb 
	 * connection */
	if (!x25_link){
		x25_link = x25_create_link(lapb_dev);
		if (!x25_link){
			printk(KERN_INFO "X25 register: Failed to create X25 link\n");
			return -ENOMEM;
		}

#ifdef CONFIG_PROC_FS
		if (x25_proc_link_init (x25_link) != X25_OK){
			x25_free_link(x25_link);
			printk(KERN_INFO "X25 register: Failed to create X25 link proc entry\n");
			return -EFAULT;
		}
#endif
		x25_insert_link(x25_link);

		X25_HOLD(x25_link);
		*x25_lnk = (void*)x25_link;
		printk(KERN_INFO "%s: Registering X25 Link\n",x25_link->name);
	
		
		if (dev_name == NULL){
			return 0;
		}	
		
	}else{
		if (x25_link_ok(x25_link) != X25_OK){
			printk(KERN_INFO "X25: Invalid X25 Link ... %s\n",
					dev_name);
			return -EINVAL;
		}
		
		if (dev_name == NULL){
			printk(KERN_INFO "%s: X25: Invalid device name : NULL!\n",
					x25_link->name);
			return -EINVAL;
		}

		if (x25_dev_in_link(x25_link,dev_name) != NULL){
			printk(KERN_INFO "X25: Device already registerd for %s\n",
					dev_name);
			return -EEXIST;
		}

	}

	if (!new_dev){
		printk(KERN_INFO "%s: X25 Failed to create interface no new_dev ptr!\n",
				x25_link->name);
		return -EFAULT;
	}	
	
	if ((x25_lcn = x25_create_lcn()) == NULL){
		printk(KERN_INFO "X25: Failed to create cb %s\n",
				dev_name);
		return -ENOMEM;
	}

	if ((dev = x25_create_dev(dev_name)) == NULL){
		printk(KERN_INFO "X25: Failed to create dev %s\n",
				dev_name);
		KFREE(x25_lcn);
		return -ENOMEM;
	}

	sprintf(x25_lcn->name,"lcn-%s",dev_name);
	dev->init = &x25_if_init;
	
	X25_HOLD(x25_lcn);
	wan_netif_set_priv(dev, x25_lcn);

	/* From this point forward, x25_free_dev() will deallocate
	 * both dev and x25_lcn */

	rtnl_lock();
	if ((err=register_netdevice(dev)) != 0){
		rtnl_unlock();
		x25_free_dev(dev);
		printk(KERN_INFO "X25: Failed to register net dev: %s  err=%i\n",
				dev_name,err);
		return -EFAULT;
	}
	rtnl_unlock();

	x25_lcn->tx_dev = lapb_dev;
	set_bit(LCN_TX_DEV,&x25_lcn->used);

	X25_HOLD(dev);
	x25_lcn->dev = dev;
	set_bit(LCN_DEV,&x25_lcn->used);

	X25_HOLD(x25_link);
	x25_lcn->x25_link = x25_link;
	set_bit(LCN_X25_LINK,&x25_lcn->used);

	
#ifdef CONFIG_PROC_FS
	if (x25_proc_add (x25_lcn, x25_lcn->x25_link->proc_dir) != X25_OK){
		unregister_netdev(dev);
#ifdef LINUX_2_6
      		wan_netif_free(dev);
#endif
		printk(KERN_INFO "X25: Failed to register proc for dev %s\n",
				dev_name);
		return -EFAULT;
	}
#endif
	
	x25_insert_dev(x25_link,dev);

	*new_dev=dev;
	return 0;
}

/*==============================================================
 * x25_unregister
 *
 * Description:
 *	This function is called during system setup to 
 *	remove the x25 link and each x25 svc defined
 *	in the wanpipe configuration file.
 *
 *	For each x25 link and x25 svc the proc file
 *	entry is removed.
 *	
 * Usedby:
 * 	Lapb layer. 
 */

int x25_unregister(struct net_device *x25_dev)
{
	x25_lcn_t *x25_lcn = dev_to_x25(x25_dev);
	x25_link_t *x25_link = NULL;
	int ret=0;
	unsigned long flags;

	if (!x25_lcn)
		return -ENODEV;
	
	if (x25_dev->flags & IFF_UP){
		printk(KERN_INFO "%s: Failed to unregister: Device UP!\n",
				x25_dev->name);
		return -EBUSY;
	}

	x25_link = x25_lcn->x25_link;
	
	if (x25_link_ok(x25_link)!=X25_OK){
		printk(KERN_INFO "%s: Failed to unregister: no link device\n",
				x25_dev->name);
		return -ENODEV;
	}

	if (x25_proc_usage_check()){
		printk(KERN_INFO "%s: Error: X25 proc directory in use!\n",
				x25_dev->name);
		return -EBUSY;
	}

	if (x25_lcn->dsp_id && 
	    IS_FUNC_CALL(x25_dsp_protocol, dsp_unregister)) {
		ret = x25_dsp_protocol.dsp_unregister(NULL, x25_lcn->dsp_id);
		if (ret != 0){
			return ret;
		}

		/* dsp_unregister() function should remove the dsp_id
		 * pointer.  This is just a precaution */
		if (x25_lcn->dsp_id){
			unsigned long flags;
			printk(KERN_INFO "%s: Warning: (x25_unregister) DSP unregister failed to remove dsp_id\n",
					x25_lcn->x25_link->name);
			spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
			x25_lcn->dsp_id = NULL;
			spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
		}
	}

	
	set_bit(LCN_UNREGISTER,&x25_lcn->critical);

#ifdef CONFIG_PROC_FS
	x25_proc_delete(x25_lcn,x25_lcn->x25_link->proc_dir);
#endif
	
	/* Schedule so to let any other processes, like tasks
	 * stop sending on this lcn, due to the critical flag
	 * set above */
	schedule();

	printk(KERN_INFO "%s: Unregistering X25 device\n",x25_dev->name);

	if (x25_lcn->x25_sw_lcn){
		printk(KERN_INFO "%s: X25 unregister disconnecting switch dev!\n",
					x25_lcn->dev->name);
		x25_sw_disconnect(x25_lcn->x25_sw_lcn);

		/* Schedule, and let the switch packet,
		 * send data out the port */
		schedule();
	}

	/* Starting critical area, do not insert any
	 * code here that can perform any kind of a wait !!!*/
	
	spin_lock_irqsave(&x25_link->bh_lock,flags);

	/* Link tx_dev is used to fairly send data over all
	 * active LCN's.  Thus its pointer could be the device
	 * that is currently going down, for sanity
	 * sake set it to NULL, and let the send bh, start
	 * sending from begining */
	if (x25_link->tx_dev){
		printk(KERN_INFO "%s: X25 unregister removing Link tx dev!\n",
					x25_lcn->dev->name);
		X25_PUT(x25_link->tx_dev);
		x25_link->tx_dev=NULL;
	}
	
	if (x25_lcn->tx_dev){
		clear_bit(LCN_TX_DEV,&x25_lcn->used);
		x25_lcn->tx_dev = NULL;
	}

	/* This is the last resort, if the x25 switch
	 * device is still bound in.  Cut the link,
	 * but clearing the pointers and releasing devices,
	 * once the switch realises that pointers are
	 * no longer linked the call will be cleared */
	{
		x25_lcn_t *x25_sw_lcn;
		if ((x25_sw_lcn=(x25_lcn_t *)x25_lcn->x25_sw_lcn) != NULL){
		
			printk(KERN_INFO "%s: X25 unregister breaking the switch link!\n",
					x25_lcn->dev->name);
			
			if (x25_sw_lcn->x25_sw_lcn){
				X25_PUT((x25_lcn_t*)x25_sw_lcn->x25_sw_lcn);
				x25_sw_lcn->x25_sw_lcn=NULL;
			}
	
			X25_PUT(x25_sw_lcn);
			x25_lcn->x25_sw_lcn=NULL;
		}
	}

	/* If the sk_id is still bound in, clear it.  This should
	 * never happen */
	if (x25_lcn->sk_id){
		printk(KERN_INFO "%s: X25 unregister breaking the sk_id link!\n",
					x25_lcn->dev->name);
		clear_bit(LCN_SK_ID,&x25_lcn->used);
		__sock_put(x25_lcn->sk_id);
		x25_lcn->sk_id=NULL;
	}
	
	x25_initialize_lcn(x25_lcn,X25_ST_HDLC_NOT_IN_ABM,HARD_INIT);

	spin_unlock_irqrestore(&x25_link->bh_lock,flags);
	

	/* This is not very prety. The refcnt must be 1
	 * so we must release the dev count since x25_link->dev
	 * has the dev pointer. However, theoretically until
	 * the dev is removed from the list the x25_lcn
	 * can still get referenced, thus I will release
	 * dev count, but still keep the pointer, until
	 * the dev has been removed.  The x25_lcn->dev
	 * pointer will be cleared in x25_free_dev() */
	if (x25_lcn->dev){
		X25_PUT(x25_lcn->dev);
	}

	x25_remove_dev(x25_link,x25_dev);

	x25_free_dev(x25_dev);

	ret = x25_link->dev_cnt;

	return ret;
}

/*==============================================================
 * x25_unregister_link
 *
 * Description:
 *	This function is called during system setup to 
 *	remove the whole x25 link and all x25 svc defined
 *	within the x25 link.
 *
 *	For each x25 link and x25 svc the proc file
 *	entry is removed.
 *	
 * Usedby:
 * 	Lapb layer. 
 */

int x25_unregister_link(void *x25_id)
{
	x25_link_t *x25_link = (x25_link_t*)x25_id;
	x25_link_t *x25_tmp;
	struct net_device *dev = NULL, *tmp_dev = NULL;
	unsigned long flags;
	int err;
	
	if (!x25_id){
		return -ENODEV;
	}
	
	WP_READ_LOCK(&x25_link_lock,flags);
	for (x25_tmp = x25_link_list; x25_tmp; x25_tmp=x25_tmp->next){
		if (x25_tmp == x25_link)
			break;
	}
	WP_READ_UNLOCK(&x25_link_lock,flags);
	
	if (!x25_tmp)
		return -ENODEV;
	
	set_bit(LINK_DOWN,&x25_link->tq_working);
	tasklet_kill(&x25_link->x25_task);
	schedule();
	
	for (dev=x25_link->dev_list; dev && *(struct net_device**)dev->priv;){
		
	 	tmp_dev = *(struct net_device**)dev->priv;
		
		err=x25_unregister(dev);
		if (err<0){
			clear_bit(LINK_DOWN,&x25_link->tq_working);
			tasklet_init(&x25_link->x25_task,x25_bh_link_send,(unsigned long)x25_link);
			x25_trigger_link_bh(x25_link,0);
			return err;
		}
		dev = tmp_dev;
	}

	/* This is the last device */
	if (dev){
		err=x25_unregister(dev);
		if (err<0){
			clear_bit(LINK_DOWN,&x25_link->tq_working);
			tasklet_init(&x25_link->x25_task,x25_bh_link_send,(unsigned long)x25_link);
			x25_trigger_link_bh(x25_link,0);
			return err;
		}
	}

#ifdef CONFIG_PROC_FS
	x25_proc_link_cleanup(x25_link);
#endif
	
	spin_lock_irqsave(&x25_link->bh_lock,flags);
	
	if (x25_link->api_sk_id){
		/* printk(KERN_INFO "WANPIPE API LISTEN STOP\n"); */
		wanpipe_api_listen_rx(NULL,x25_link->api_sk_id);		
		__sock_put(x25_link->api_sk_id);
		x25_link->api_sk_id = NULL;
	}
	
	if (x25_link->dsp_sk_id){
		/* printk(KERN_INFO "WANPIPE DSP LISTEN STOP\n"); */
		wanpipe_api_listen_rx(NULL,x25_link->dsp_sk_id);
		__sock_put(x25_link->dsp_sk_id);
		x25_link->dsp_sk_id = NULL;
	}

	x25_link->state = X25_ST_HDLC_NOT_IN_ABM;
	x25_clear_link_queues(x25_link,X25_ST_HDLC_NOT_IN_ABM);

	spin_unlock_irqrestore(&x25_link->bh_lock,flags);

	
	lapb_link_unregister(x25_link->master);
	X25_PUT(x25_link);
	x25_link->master=NULL;

	x25_remove_link(x25_link);
	x25_free_link(x25_link);
	return 0;
}

/* ====================================================
 * x25_dsp_id_unregister
 *
 * Used by upper layer (DSP) to remove the
 * dsp_id link from the x25 layer.
 *
 * Return: 0 = ready
 * 	   non-zero = not ready
 */

void x25_dsp_id_unregister(struct net_device *dev)
{
	x25_lcn_t *x25_lcn=dev_to_x25(dev);
	
	if (!dev){
		return;
	}
	
	if (x25_lcn == NULL){
		return;
	}

	if (x25_lcn->dsp_id){
		unsigned long flags;

		printk(KERN_INFO "%s: X25 Unregistering dsp id\n",
				x25_lcn->dev->name);

		spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
		x25_lcn->dsp_id=NULL;
		spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
	}
	return;	
}

/* ====================================================
 * x25_dsp_is_svc_ready
 *
 * Used by upper layer (DSP) to check if the
 * interface is in ready state 
 *
 * Return: 0 = ready
 * 	   non-zero = not ready
 */

int x25_dsp_is_svc_ready(struct net_device *dev)
{
	x25_lcn_t *x25_lcn=dev_to_x25(dev);

	if (!x25_lcn){
		return -ENODEV;
	}

	if (x25_lcn->state == X25_ST_PKT_LVL_RDY_P1){
		return 0;
	}

	return 1;
}

/*==============================================================
 * x25_setparms
 *
 * Description:
 *	This function is called during system setup to 
 *	configure the x25 link with the x25 profile
 *	defined in wanpipe configuration file.
 *
 *	As well it is used to configure each svc within
 *	the x25 link.
 *
 * Usedby:
 * 	Configurion layer (wanmain.c) 
 */

int x25_setparms(struct net_device *dev, struct x25_parms_struct *conf)
{
	x25_lcn_t *x25_lcn = dev_to_x25(dev);
	x25_link_t *x25_link;
	
	if (x25_lcn == NULL){
		printk(KERN_INFO "%s: Private SVC structure not found!\n",
				dev->name);
		return -EINVAL;
	}

	x25_link = x25_lcn->x25_link;

	x25_link->cfg.X25_API_options 			= conf->X25_API_options;
	x25_link->cfg.X25_protocol_options 		= conf->X25_protocol_options;
	x25_link->cfg.X25_response_options		= conf->X25_response_options;
	x25_link->cfg.X25_statistics_options 		= conf->X25_statistics_options;
	
	x25_link->cfg.packet_window			= conf->packet_window;
	x25_link->cfg.default_packet_size		= conf->default_packet_size;
	x25_link->cfg.maximum_packet_size		= conf->maximum_packet_size; 
	
	x25_link->cfg.highest_PVC			= conf->highest_PVC;	
	x25_link->cfg.lowest_PVC			= conf->lowest_PVC;
	x25_link->cfg.lowest_two_way_channel		= conf->lowest_two_way_channel;	
	x25_link->cfg.highest_two_way_channel		= conf->highest_two_way_channel;


	if (x25_link->cfg.highest_PVC && x25_link->cfg.highest_two_way_channel){
		if (x25_link->cfg.highest_PVC >= x25_link->cfg.lowest_two_way_channel){
			printk("%s: Error: Highest PVC (%u) > Lowest SVC (%u)!\n",
					x25_link->name,
					x25_link->cfg.highest_PVC,
					x25_link->cfg.lowest_two_way_channel);
			return -EINVAL;
		}	
	}		
	
	x25_link->cfg.lowest_incoming_channel		= conf->lowest_incoming_channel;	
	x25_link->cfg.highest_incoming_channel		= conf->highest_incoming_channel;
	x25_link->cfg.lowest_outgoing_channel		= conf->lowest_outgoing_channel;	
	x25_link->cfg.highest_outgoing_channel		= conf->highest_outgoing_channel;


	x25_link->cfg.genl_facilities_supported_1	= conf->genl_facilities_supported_1;	
	x25_link->cfg.genl_facilities_supported_2	= conf->genl_facilities_supported_2;	
	x25_link->cfg.CCITT_facilities_supported	= conf->CCITT_facilities_supported;	
	x25_link->cfg.non_X25_facilities_supported	= conf->non_X25_facilities_supported;	
	
	x25_link->cfg.CCITT_compatibility		= conf->CCITT_compatibility;				
	/* By shifting right 10 bits, its equivalent of *1024 Multiplication */
	
	x25_link->cfg.T10_T20				= conf->T10_T20*HZ;	
	x25_link->cfg.T11_T21				= conf->T11_T21*HZ;
	x25_link->cfg.T12_T22				= conf->T12_T22*HZ;
	x25_link->cfg.T13_T23				= conf->T13_T23*HZ;	
	x25_link->cfg.T16_T26				= conf->T16_T26*HZ;	
	x25_link->cfg.T28				= conf->T28*HZ;

	x25_link->cfg.R10_R20				= conf->R10_R20;
	x25_link->cfg.R12_R22				= conf->R12_R22;	
	x25_link->cfg.R13_R23				= conf->R13_R23;		


	DEBUG_TEST("%s: X25 Link: Win %i MTU %i MaxMTU %i LSVC %i HSVC %i\n",
			x25_link->name,
			x25_link->cfg.packet_window,
			x25_link->cfg.default_packet_size,
			x25_link->cfg.maximum_packet_size,
			x25_link->cfg.lowest_two_way_channel,
			x25_link->cfg.highest_two_way_channel);

	
	if (x25_link->cfg.call_logging != conf->call_logging){
		x25_link->cfg.call_logging = conf->call_logging;
		if (!x25_link->cfg.call_logging){
			printk(KERN_INFO "%s: Disabling Call Logging\n",
					x25_link->name);
		}else{
			printk(KERN_INFO "%s: Enabling Call Logging\n",
					x25_link->name);
		}
	}
	
	x25_lcn->R12_R22 = x25_link->cfg.R12_R22;
	x25_lcn->R13_R23 = x25_link->cfg.R13_R23;
	
	x25_link->cfg.dte = conf->dte;
	
	switch (conf->mode){

	case MOD8_MODULUS:
		x25_link->cfg.mode = GFI_FOR_MOD_8;
		if (x25_link->cfg.packet_window < 1 || x25_link->cfg.packet_window > 7){
			DEBUG_EVENT("%s: Error: Invalid Packet Window %i! MOD8 Win 1-7\n",
					x25_link->name,x25_link->cfg.packet_window);
			return -EINVAL;
		}
		break;
		
	case MOD128_MODULUS:
		x25_link->cfg.mode = GFI_FOR_MOD_128;
		if (x25_link->cfg.packet_window < 1 || x25_link->cfg.packet_window > 127){
			DEBUG_EVENT("%s: Error: Invalid Packet Window %i! MOD128 Win 1-127\n",
					x25_link->name,x25_link->cfg.packet_window);
			return -EINVAL;
		}
		break;
	default:
		printk(KERN_INFO "%s: ERROR: Invalid X25 Modulus %i\n",
				dev->name, conf->mode);
		return -EINVAL;
	}


	x25_link->incoming_call_handling = ACCEPT_ALL_INCOMING_CALLS;

	/* SVC Configuration */
	x25_lcn->rx_window_size = x25_link->cfg.packet_window;
	x25_lcn->rx_pkt_size =	x25_link->cfg.default_packet_size; 
		
	x25_lcn->tx_window_size = x25_link->cfg.packet_window;
	x25_lcn->tx_pkt_size = x25_link->cfg.default_packet_size;

	x25_lcn->call_backoff_timeout = conf->call_backoff_timeout*HZ;
	
#if defined(X25_MTU_CHOP)
	dev->mtu=4096;
#else		
	dev->mtu=x25_lcn->tx_pkt_size;	
#endif
	lapb_set_mtu(x25_link->master, (x25_link->cfg.maximum_packet_size + 
				         ((x25_link->cfg.mode == GFI_FOR_MOD_8) ? 3 : 4)));
	
	strncpy(x25_lcn->call_string,  conf->call_string, X25_CALL_STR_SZ);
	strncpy(x25_lcn->accept_called, conf->accept_called, WAN_ADDRESS_SZ);
	strncpy(x25_lcn->accept_calling, conf->accept_calling, WAN_ADDRESS_SZ);
	strncpy(x25_lcn->accept_facil, conf->accept_facil, WAN_ADDRESS_SZ);
	strncpy(x25_lcn->accept_udata, conf->accept_udata, WAN_ADDRESS_SZ);
	strncpy(x25_lcn->label,conf->label,WAN_IF_LABEL_SZ);

	memset(x25_lcn->dlabel,0,WAN_IF_LABEL_SZ);
	
	/* X25_SW X.25 switching */
	strncpy(x25_link->virtual_addr, conf->virtual_addr, WAN_ADDRESS_SZ);
	strncpy(x25_link->real_addr, conf->real_addr, WAN_ADDRESS_SZ);

	DEBUG_EVENT("%s: Vir %s Real %s Label %s PVC %s\n",
				x25_link->name,
				x25_link->virtual_addr,
				x25_link->real_addr,
				x25_lcn->label,
				is_digit(conf->addr[0])?"YES":"NO");

	/* PVC */
	if (is_digit(conf->addr[0])){	/* PVC */
		int lcn = dec_to_uint(conf->addr, 0);
		int err;

		DEBUG_EVENT("%s: Configuring for PVC %i\n",dev->name,lcn);

		if ((lcn >= x25_link->cfg.lowest_PVC) && (lcn <= x25_link->cfg.highest_PVC)){

			atomic_set(&x25_lcn->lcn,lcn);	
			err=x25_bind_lcn_to_dev(x25_link,x25_lcn);
			if (err){
				printk(KERN_INFO "%s: PVC %u busy!\n",
						dev->name,lcn);
				atomic_set(&x25_lcn->lcn,0);	
				return err;
			}
			set_bit(LCN_USED,&x25_lcn->used);
			set_bit(LCN_CONFIG_PVC,&x25_lcn->lcn_config);

		}else{
			DEBUG_EVENT("%s: PVC %u is out of range (Min=%u Max=%u)!\n",
				dev->name, lcn, x25_link->cfg.lowest_PVC,x25_link->cfg.highest_PVC);
			return -EINVAL;
		}
	}

	return X25_OK;
}



/***************************************************************

 *	SVC NETWORK INTERFACE DEVICE FUNCTIONS 
 
 ***************************************************************/





/*==============================================================
 * if_tx_timeout
 *
 * Description:
 * 	Network device timeout function called by the
 * 	TCP/IP stack if the device is shutdown for a 
 * 	long time. 
 *
 * 	Only used in TCP/IP mode (not in API).
 * 
 */

static void if_tx_timeout (struct net_device *dev)
{
	/* If our device stays busy for at least 5 seconds then we will
	 * kick start the device by making dev->tbusy = 0.  We expect
	 * that our device never stays busy more than 5 seconds. So this                 
	 * is only used as a last resort.
	 */

	printk (KERN_INFO "%s: Transmit timed out.\n", 
			dev->name);
	dev->trans_start = jiffies;
	netif_wake_queue (dev);

}

/*==============================================================
 * x25_ioctl
 *
 * Description:
 *	Network device ioctl function used by the socket api 
 *	layer or higher layers, to execute X25 commands and 
 *	perform registration and unregistration tasts.
 *
 * Usedby:
 * 	Higher protocol layers such as socket api, dsp, etc.
 */

int x25_ioctl (struct net_device *dev, struct ifreq *ifr, int cmd)
{
	x25_lcn_t *x25_lcn=dev_to_x25(dev);
	int err=0;
	unsigned long flags;

	if (!dev || !(dev->flags & IFF_UP)){
		return -ENODEV;
	}
	
	if (x25_lcn == NULL){
		return -ENODEV;
	}

	if (test_bit(LINK_DOWN,&x25_lcn->x25_link->tq_working)){
		printk(KERN_INFO "%s: X25 Ioctl failed, x25 link down\n",x25_lcn->x25_link->name);
		return -EINVAL;
	}

	if (test_and_set_bit(IOCTL_SMP,&x25_lcn->critical)){
		printk(KERN_INFO "%s: CRITICAL IN IOCTL Cmd=0x%X CHECK_TX 0x%X\n",
				x25_lcn->dev->name,cmd,
				SIOC_WANPIPE_CHECK_TX);
		return -EINVAL;
	}

	switch (cmd){
	
		case SIOC_X25_PLACE_CALL:
			err=x25_api_place_call(x25_lcn,ifr);
			break;
			
		case SIOC_X25_SET_CALL_DATA:
			err=x25_api_cmd_setup(x25_lcn,ifr);
			break;
			
		case SIOC_X25_GET_CALL_DATA:
			err=x25_api_get_cmd_result(x25_lcn, ifr);
			break;
			
		case SIOC_X25_CLEAR_CALL:

			if (IS_PVC(x25_lcn)){
				err=0;
				goto x25_ioctl_exit;
			}
			
			err=x25_api_clear_call(x25_lcn,ifr);
			break;
		
		case SIOC_X25_ACCEPT_CALL:
			err=x25_api_accept_call(x25_lcn,ifr);
			break;
		
		case SIOC_X25_RESET_CALL:
			err=x25_api_reset_call(x25_lcn,ifr);
			break;
			
		case SIOC_X25_HDLC_LINK_STATUS:
			DEBUG_EVENT("%s: HDLC LINK STATUS %d\n",
					x25_lcn->x25_link->name, x25_lcn->x25_link->state);
			err= x25_lcn->x25_link->state == X25_ST_PKT_LVL_RDY_R1 ? 1 : 0;
			break;

		case SIOC_X25_SET_LCN_PID:
			if (ifr && ifr->ifr_data){
				err = copy_from_user(&x25_lcn->pid,
				  	              ifr->ifr_data,
					              sizeof(pid_t));
			}
			break;

		case SIOC_X25_SET_LCN_LABEL:
			if (ifr && ifr->ifr_data){
				err = copy_from_user(&x25_lcn->dlabel[0],
				  	              ifr->ifr_data,
						      WAN_IF_LABEL_SZ);

				x25_lcn->dlabel[WAN_IF_LABEL_SZ]=0;
				DEBUG_TEST("%s: Setting DLABLE to %s\n",
						x25_lcn->dev->name,x25_lcn->dlabel);
			}
			break;	
			
		case SIOC_ANNEXG_BIND_SK:
			if (!ifr){
				err= -EINVAL;
				goto x25_ioctl_exit;
			}
				
			if (ifr->ifr_data && !x25_lcn->sk_id){
				if (!spin_is_locked(&x25_lcn->x25_link->bh_lock)){
					spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
					x25_lcn->sk_id = ifr->ifr_data;
					sock_hold(x25_lcn->sk_id);
					set_bit(LCN_SK_ID,&x25_lcn->used);
					spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
				}else{
					x25_lcn->sk_id = ifr->ifr_data;
					sock_hold(x25_lcn->sk_id);
					set_bit(LCN_SK_ID,&x25_lcn->used);
				}
				break;
			}
		
			err= -EBUSY;
			break;

		case SIOC_ANNEXG_UNBIND_SK:
			if (!ifr){
				err= -EINVAL;
				goto x25_ioctl_exit;
			}

			if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){

				if (ifr->ifr_data != x25_lcn->sk_id){
					printk(KERN_INFO "%s: ERROR: API trying to unbind invlaid sock api! New=%p Orig=%p\n",
							x25_lcn->dev->name,ifr->ifr_data,x25_lcn->sk_id);
					err=-EINVAL;
					goto x25_ioctl_exit;
				}
			
				if (!spin_is_locked(&x25_lcn->x25_link->bh_lock)){
					spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
					clear_bit(LCN_SK_ID,&x25_lcn->used);
					__sock_put(x25_lcn->sk_id);
					x25_lcn->sk_id=0;
					spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
				}else{
					clear_bit(LCN_SK_ID,&x25_lcn->used);
					__sock_put(x25_lcn->sk_id);
					x25_lcn->sk_id=0;
				}

				if (IS_PVC(x25_lcn)){
					x25_lcn->dlabel[0]=0;
					x25_lcn->pid=0;

					/* The PVC can be only in two states,
					 * READY or RESET SENT. If we have sent a reset
					 * and did not receive a reset conf, there are
					 * two possibilites.
					 *
					 * 1: The other end is up and well get and rest
					 *    conf which will set the x25_lcn state to
					 *    Ready.
					 *
					 * 2: The other end is down, and it will not
					 *    respond to our reset.  Thus on exit our
					 *    state will stay in reset sent mode.
					 *    In this case we should stop the reset 
					 *    timeout function, so it doesn't keep
					 *    sending the reset, once the app has
					 *    closed 
					 */
					if (x25_lcn->state != X25_ST_PKT_LVL_RDY_D1){
						spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
						x25_initialize_lcn(x25_lcn,
								   x25_lcn->state,
								   SOFT_INIT);
						spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
					}
				}

				break;
			}
		
			err=-ENODEV;
			break;

		case SIOC_ANNEXG_GET_SK:
			if (!ifr){
				err= -EINVAL;
				goto x25_ioctl_exit;
			}
			
			if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
				sock_hold(x25_lcn->sk_id);
				ifr->ifr_data = (void*)x25_lcn->sk_id;
			}else{
				ifr->ifr_data = NULL;
			}
			break;
		
		case SIOC_WANPIPE_CHECK_TX:

			if (skb_queue_len(&x25_lcn->write_queue)){
				err=1;
			}else{
				if (test_bit(LCN_SENDING,&x25_lcn->critical)){
					err=1;
				}
			}
			break;

		case SIOC_ANNEXG_KICK:
			x25_trigger_link_bh(x25_lcn->x25_link,1);
			if (test_bit(LCN_TX_DEV,&x25_lcn->used) && x25_lcn->tx_dev){
				lapb_kickstart_tx(x25_lcn->tx_dev);
			}
			break;

		case SIOC_WANPIPE_SNMP:
			err= x25_snmp_data(x25_lcn, ifr);
			break;

		default:
			printk(KERN_INFO "%s: Unsupported Ioctl Call 0x%x \n",
					x25_lcn->dev->name,cmd);
			err= -EOPNOTSUPP;
			break;
	}
	
x25_ioctl_exit:

	clear_bit(IOCTL_SMP,&x25_lcn->critical);
	return err;	
}

int x25_set_config (struct net_device *dev,struct ifmap *map)
{
	return 0;	
}

/*==============================================================
 * x25_open_dev
 *
 * Description:
 *	This function enables the svc network device, and
 *	sets it up for data transfer.
 *
 *	It is called by the kernel during interfaces
 *	startup via ifconfig:
 *		ex: ifconfig wp1_svc up
 *
 *  	If the network device contains IP data, its
 *  	operation mode is set to TCP/IP otherwise,
 *  	the opteration mode is API.
 * 
 * Usedby:
 * 	Kernel during ifconfig system call.
 */

int x25_open_dev(struct net_device *dev)
{
	struct in_ifaddr *ifaddr;
	struct in_device *in_dev;
	x25_lcn_t *x25_lcn = dev_to_x25(dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
	if (netif_running(dev))
		return -EBUSY;
#endif

	clear_bit(LCN_DOWN,&x25_lcn->critical);
	clear_bit(0,&x25_lcn->api_state);
	
	if (x25_lcn->x25_link->state == X25_ST_PKT_LVL_RDY_R1){
		lcn_state_change(x25_lcn,X25_ST_PKT_LVL_RDY_P1);
		netif_start_queue(dev);
		if (x25_lcn->dsp_id){
			x25_update_dsp_state(x25_lcn);	
		}
	}

	if ((in_dev = in_dev_get(dev)) == NULL){
		x25_lcn->usedby = API;
#if defined(X25_MTU_CHOP)
		dev->mtu += sizeof(x25api_hdr_t);
#else
		dev->mtu = x25_lcn->tx_pkt_size+sizeof(x25api_hdr_t);
#endif
		dev->tx_timeout	= NULL;		
		return 0;
	}

	if ((ifaddr = in_dev->ifa_list)== NULL ){
		x25_lcn->usedby = API;
		dev->mtu = x25_lcn->tx_pkt_size+sizeof(x25api_hdr_t);
		dev->tx_timeout	= NULL;		
		return 0;
	}else{
		x25_lcn->usedby = TCPIP;
	}

	return 0;
}


/*==============================================================
 * x25_stop_dev
 *
 * Description:
 *	This function disables the svc network device.
 *
 *	It is called by the kernel during interfaces
 *	shutdown via ifconfig:
 *		ex: ifconfig wp1_svc down
 *
 *	If the svc is in connected state, the call
 *	will be disconnected.
 *	
 * Usedby:
 * 	Kernel during ifconfig system call.
 */

int x25_stop_dev(struct net_device *dev)
{
	x25_lcn_t *x25_lcn = dev_to_x25(dev);

	netif_stop_queue(dev);

	if (!x25_lcn)
		return 0;

	set_bit(LCN_DOWN,&x25_lcn->critical);

	if (atomic_read(&x25_lcn->lcn)){
		unsigned long flags;
		x25_lcn_t *x25_sw_lcn;
		
		printk(KERN_INFO "%s: Interface down, releaseing SVC!\n",
				dev->name);

		if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
			spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
			clear_bit(LCN_SK_ID,&x25_lcn->used);
			__sock_put(x25_lcn->sk_id);
			x25_lcn->sk_id=0;
			spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
		}
		
		if ((x25_sw_lcn=x25_lcn->x25_sw_lcn) != NULL){
			spin_lock_irqsave(&x25_sw_lcn->x25_link->bh_lock,flags);
			x25_sw_disconnect(x25_sw_lcn);
			spin_unlock_irqrestore(&x25_sw_lcn->x25_link->bh_lock,flags);
		}

		spin_lock_irqsave(&x25_lcn->x25_link->bh_lock,flags);
		x25_clear_call(x25_lcn,1);
		spin_unlock_irqrestore(&x25_lcn->x25_link->bh_lock,flags);
	}

	clear_bit(0,&x25_lcn->api_state);
	return 0;
}

/*==============================================================
 * x25_ifstats
 *
 * Description:
 * 	This fucntion interfaces the /proc file system
 * 	to the svc device.  The svc keeps protocol and
 * 	packet statistcs.  This function passes these
 * 	statistics to the /proc file system.
 *	
 * Usedby:
 * 	Kernel /proc file system
 * 	x25_proc.c function also uses it to obtain svc
 * 	statistics.
 */

static struct net_device_stats gstats;
static struct net_device_stats* x25_ifstats (struct net_device *dev)
{
	x25_lcn_t *x25_lcn=dev_to_x25(dev);

	if (!dev || !(dev->flags&IFF_UP))
		return &gstats;

	if (x25_lcn == NULL)
		return &gstats;

	return &x25_lcn->ifstats;
}

/*==============================================================
 * x25_if_init
 *
 * Description:
 *	During device registration, this function is
 *	called to fill in the call back functions
 *	used in nework device setup and operation.
 *	
 *	The kernel interfaces the driver, using the
 *	call back functions below.
 *	
 * Usedby:
 * 	Kernel during register_netdevice() in x25_register
 * 	function.
 */

static int x25_if_init(struct net_device *dev)
{
	dev->open		= &x25_open_dev;
	dev->stop		= &x25_stop_dev;
	dev->hard_header	= NULL;
	dev->rebuild_header	= NULL;
	dev->hard_start_xmit	= &x25_if_send;
	dev->get_stats		= &x25_ifstats;
	dev->tx_timeout		= &if_tx_timeout;
	dev->watchdog_timeo	= 10;
	dev->hard_header_len	= 25;
	dev->do_ioctl		= &x25_ioctl;
	dev->set_config		= &x25_set_config;
	
		/* Initialize media-specific parameters */
	dev->flags		|= IFF_POINTOPOINT;
	dev->flags		|= IFF_NOARP;

	dev->tx_queue_len = 10;
	dev->type = ARPHRD_X25;
	dev->addr_len = 0;
	*(u8*)dev->dev_addr = 0; 

	/* Initialize socket buffers */
	dev_init_buffers(dev);
	return 0;
}

/*==============================================================
 * x25_if_send
 *
 * Description:
 * 	Call back function used by the kernel or the
 * 	upper protocol layer to transmit data for each
 * 	svc withing the x25 link.
 *	
 *	Data can only be tranmsitted if the svc state
 *	is connected.
 *
 *	If state != CONNECTED && If svc mode=TCP/IP
 *		place an x25 call and try to establish
 *		connection.
 *		
 *	If state != CONNECTED && If svc mode=API
 *		refuse the packet, and indicate to 
 *		the upper layer that the connection has
 *		not been made.
 * 	
 *	
 * Usedby:
 * 	Kernel TCP/IP stack or upper layers to transmit data.
 */

static int x25_if_send (struct sk_buff* skb, struct net_device* dev)
{
	x25_lcn_t *x25_lcn = dev_to_x25(dev);
	int data_len = 0;
	x25api_hdr_t *x25_hdr;

	/* This  is a sanity check */
	if (!x25_lcn || !(dev->flags&IFF_UP)){
		printk(KERN_INFO "x25_if_send: Major error no device!\n");
		if (skb){
			X25_SKB_INC(skb->truesize);
			KFREE_SKB(skb);
		}
		return 0;
	}
	
	x25_hdr=(x25api_hdr_t *)skb->data;
	
	netif_stop_queue(dev);

	clear_bit(IF_SEND_AWAITING_KICK,&x25_lcn->critical);

	if (x25_lcn->usedby == API){
		data_len = skb->len - sizeof(x25api_hdr_t);
	}else{
		data_len = skb->len;
	}
	
	X25_SKB_INC(skb->truesize);

	
#if !defined( X25_MTU_CHOP)	
	if (data_len > x25_lcn->tx_pkt_size){
		printk(KERN_INFO "%s: Tx packet greater than MTU (%i>%i)!\n",
				x25_lcn->dev->name,data_len,x25_lcn->tx_pkt_size);
		netif_start_queue(dev);
		
		if (x25_lcn->usedby == API){
			X25_SKB_DEC(skb->truesize);
			return -EINVAL;
		}else{
			KFREE_SKB(skb);
			dev->trans_start = jiffies;
			return 0;
		}
	}
#endif

	if (x25_lcn->usedby == API && x25_hdr->qdm & 0x01 && data_len < x25_lcn->tx_pkt_size){
		printk(KERN_INFO "%s: WARNING: Tx packet Mbit set, but Tx Len (%i) < MTU (%i) \n",
				dev->name,data_len,x25_lcn->tx_pkt_size);
	}

		
	if (x25_lcn->x25_link->state != X25_ST_PKT_LVL_RDY_R1){
		printk(KERN_INFO "HDLC link not in ABM ! %i\n",x25_lcn->x25_link->state);
		netif_start_queue(dev);
		
		if (x25_lcn->usedby == API){
			X25_SKB_DEC(skb->truesize);
			return -EINVAL;
		}else{
			KFREE_SKB(skb);
			dev->trans_start = jiffies;
			return 0;
		}
	}
	
	if (x25_lcn->state != X25_ST_DATA_TRANSFER_P4){

		if (x25_lcn->usedby == TCPIP && x25_lcn->state == X25_ST_PKT_LVL_RDY_P1){
			if ((jiffies - x25_lcn->call_backoff) > x25_lcn->call_backoff_timeout){
				x25_lcn->call_backoff = jiffies;
				x25_place_call(x25_lcn);
			}
		}
		netif_start_queue(dev);

		if (x25_lcn->usedby == API){
			X25_SKB_DEC(skb->truesize);
			return -EINVAL;
		}else{
			KFREE_SKB(skb);
			dev->trans_start = jiffies;
			return 0;
		}
	}

	if (test_bit(TX_WINDOW_CLOSED,&x25_lcn->critical)){
		printk(KERN_INFO "%s: IF SEND TX Window Closed !!!\n",dev->name);
		netif_stop_queue(dev);
		set_bit(IF_SEND_AWAITING_KICK,&x25_lcn->critical);
	
		X25_SKB_DEC(skb->truesize);

		if (x25_lcn->usedby == API){
			return -EBUSY;
		}else{
			return 1;
		}
	}

	/* If our current queue already holds maximum number
	 * of unacked packets, then block the upper layer 
	 * from transmitting until the queue clears.
	 *
	 * Otherwise queue up the packet and call the
	 * x25_kick bottom half handler to transmit the
	 * packet.*/

	if (skb_queue_len(&x25_lcn->write_queue) >= x25_lcn->tx_window_size){
		DEBUG_TEST("The Queue Len is at Maximum Qlen %i Win %i!\n",
				skb_queue_len(&x25_lcn->write_queue),x25_lcn->tx_window_size);
		netif_stop_queue(dev);

		set_bit(IF_SEND_AWAITING_KICK,&x25_lcn->critical);
		x25_trigger_link_bh(x25_lcn->x25_link,0);
	
		X25_SKB_DEC(skb->truesize);

		if (x25_lcn->usedby == API){
			return -EBUSY;
		}else{
			return 1;
		}
	}

	wan_skb_unlink(skb);
	
	/* Since we need to insert x25 headers into the
	 * data received from the upper layers we must check
	 * that there is enough room to do so. If not
	 * we must make some room */
	if (skb_headroom(skb) < 16 || skb_tailroom(skb) < 16){
		struct sk_buff *skb2 = NULL;
		
		skb2 = skb_copy_expand(skb, 16, 16, GFP_ATOMIC);
		
		if (skb2 == NULL){
			printk(KERN_INFO "%s: Error in if_send: head reallocatin pkt\n",
					dev->name);
			dev->trans_start = jiffies;
			netif_start_queue(dev);
			KFREE_SKB(skb);
			return 0;
		}

		KFREE_SKB(skb);
		X25_SKB_INC(skb2->truesize);
		skb = skb2;	
		wan_skb_unlink(skb);
	}
	
	/* Queue the incoming packet into the svc write
	 * queue, and trigger the x25 link TX Bottom Half
	 * queue.  We can only transmit if no async 
	 * data is present and only if our state is 
	 * CONNECTED */

	skb_queue_tail(&x25_lcn->write_queue, skb);

	x25_trigger_link_bh(x25_lcn->x25_link,0);
		
	dev->trans_start = jiffies;
	netif_start_queue(dev);
	return 0;
}


/***************************************************************
 *
 * 	INTERFACE FUNCTION BETWEEN LAPB and X25 LAYER
 * 
 ***************************************************************/




/*==============================================================
 * x25_receive_pkt
 *
 * Description:
 *
 * 	Called by the labp protocol to give us the packet.
 * 	The packet is passed throught the x25 state 
 * 	machine in x25_in.c
 *	
 * Usedby:
 * 	Lower layer to pass us an rx packet.
 */

int x25_receive_pkt(void *x25_id, struct sk_buff *skb)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;

	if (!skb){
		printk(KERN_INFO "X25 received data without the skb!\n");
		return X25_OK;
	}
	
	X25_SKB_INC(skb->truesize);

	wan_skb_unlink(skb);

	if (!x25_link){
		printk(KERN_INFO "X25 received data without the link id!\n"); 	
		KFREE_SKB(skb);
		return X25_OK;
	}

	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		printk(KERN_INFO "%s: Rx drop, link down!\n",x25_link->name);
		KFREE_SKB(skb);
		return X25_OK;
	}

	if (x25_link->state == X25_ST_HDLC_NOT_IN_ABM){
		printk(KERN_INFO "%s: Rx drop, X25 link not in ABM Mode\n",
				x25_link->name);
		KFREE_SKB(skb);
		return X25_OK;
	}

	if (skb_queue_len(&x25_link->rx_queue) >= MAX_LINK_RX_Q_LEN){
		printk(KERN_INFO "%s: Warning: X25 Link Rx queue full... \n",
				x25_link->name);
		X25_SKB_DEC(skb->truesize);

		/* Trigger the bh just in case */
		x25_trigger_link_bh(x25_link,1);
		return 1;
	}
	
	skb_queue_tail(&x25_link->rx_queue,skb);

	x25_trigger_link_bh(x25_link,1);
	return 0;
}

/*==============================================================
 * x25_data_tx_up
 *
 * Description:
 *	This function is used to pass data to the upper 
 *	layer. If the svc is used by TCP/IP the packet
 *	is passed up the the TCP/IP stack, otherwise
 *	it is passed to the upper protocol layer.
 *	
 * Usedby:
 * 	x25 layer to pass data up to the next layer.
 */

int x25_data_tx_up(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	int err=0;
	x25api_hdr_t *x25_api_hdr=NULL;
	int len=skb->len;
	int truesize=skb->truesize;
	
	wan_skb_unlink(skb);
	skb->protocol = htons(ETH_P_IP);
	wan_skb_reset_mac_header(skb);

	if (x25_lcn->usedby==API){
		
		skb->pkt_type = WAN_PACKET_DATA;
		
		if (skb_headroom(skb) < sizeof(x25api_hdr_t)){
			printk(KERN_INFO "NO Room for X25 Header!\n");
			goto tx_up_drop_pkt;
		}
		
		x25_api_hdr = (x25api_hdr_t*)skb_push(skb,sizeof(x25api_hdr_t));

		x25_api_hdr->qdm = (frame->gfi >> 1);
		if (frame->mbit){
			set_bit(0,(unsigned long*)&x25_api_hdr->qdm);
		}else{
			clear_bit(0,(unsigned long*)&x25_api_hdr->qdm);	
		}

		x25_api_hdr->cause = frame->cause;
		x25_api_hdr->diagn = frame->diag;
		x25_api_hdr->pktType = frame->type;
		x25_api_hdr->length  = len;
		x25_api_hdr->result = 0;
		x25_api_hdr->lcn = frame->lcn;
		
		if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
			
			if (IS_PVC(x25_lcn) && 
			    x25_lcn->state != X25_ST_DATA_TRANSFER_P4){
				/* It is possible to be in network
				 * down state for PVC.  In such case
				 * drop the packet */
				x25_lcn->ifstats.rx_dropped++;
				goto tx_up_drop_pkt;
			}
			
			err = wanpipe_api_sock_rx(skb,x25_lcn->dev,x25_lcn->sk_id);
			if (err != X25_OK){
				err=0;
				if (err == -EINVAL){
					printk(KERN_INFO "%s:Major Error in Socket Above: Tx UP!!!\n",
							x25_lcn->dev->name);
					clear_bit(LCN_SK_ID,&x25_lcn->used);
					__sock_put(x25_lcn->sk_id);
					x25_lcn->sk_id = NULL;
				}
				KFREE_SKB(skb);
				x25_clear_call(x25_lcn,1);
			}
		
		}else if (x25_lcn->dsp_id) {
			
			if (!IS_FUNC_CALL(x25_dsp_protocol, dsp_receive_pkt)){
				printk(KERN_INFO "X25 TX UP: No DSP protocol!\n");
				goto tx_up_drop_pkt;
			}

			err = x25_dsp_protocol.dsp_receive_pkt(x25_lcn->dsp_id, skb);
			if (err != X25_OK){
				KFREE_SKB(skb);
			}
		}else{
			if (IS_PVC(x25_lcn)){
				/* It is legal for pvc not to be
				 * attached to a socket.  In such
				 * case drop the packet */
				x25_lcn->ifstats.rx_dropped++;
				goto tx_up_drop_pkt;
			}
		}
		
	}else{
		skb->dev = x25_lcn->dev;
		netif_rx(skb);
	}

	if (!err){
		x25_lcn->ifstats.rx_packets++;
		x25_lcn->ifstats.rx_bytes += len;
		X25_SKB_DEC(truesize);
	}	
	
	return err;
		
tx_up_drop_pkt:
	KFREE_SKB(skb);
	return 0;
}


/*==============================================================
 * x25_data_tx_down
 *
 * Description:
 *	This function is used to pass data down to the 
 *	lower layer (Lapb). 
 *	
 */

int x25_data_tx_down(x25_link_t *x25, struct sk_buff *skb)
{
	int len;
	int err=0;

	len = skb->truesize;
	wan_skb_unlink(skb);

	if (x25->master){

		if (!(x25->master->flags&IFF_UP)){
			printk(KERN_INFO "%s: Critical error master dev %s Down: data lost!\n",x25->name,x25->master->name);
			KFREE_SKB(skb);
			return 0;
		}
		
		if (netif_queue_stopped(x25->master)){
			return -EBUSY;
		}

		err = x25->master->hard_start_xmit(skb,x25->master);

		if (!err){
			X25_SKB_DEC(len);
		}
		return err;
	}else{
		printk(KERN_INFO "X25: No Master Device!: data lost\n");
		KFREE_SKB(skb);
		return 0;
	}
}

/*==============================================================
 * x25_lapb_connect_confirm
 *
 * Description:
 * 	The lapb layer calls this function, when it
 * 	becomes connected.  Once the lapb layer
 * 	get connected, we can start the x25 protocol.
 */

void x25_lapb_connect_confirm(void *x25_id,int reason)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;
	unsigned long flags;
	
	clear_bit(BH_AWAITING_KICK,&x25_link->tq_working);

	printk(KERN_INFO "%s: X25 Link connecting (tx restart req) ... \n",
			x25_link->name);


	spin_lock_irqsave(&x25_link->bh_lock,flags);

	x25_link->state = X25_ST_RESTART_REQ_SNT_R2;
	x25_clear_link_queues(x25_link,X25_ST_RESTART_REQ_SNT_R2);

	/* Some networks don't like us to send RESATART on ABM */
	if(!(x25_link->cfg.X25_protocol_options & NO_RESTART_REQ_ON_ENTER_ABM)){
		x25_send_control(x25_link,0,0,RESTART_REQUEST_PKT,0,0,NULL,0);
		x25_start_t10_t20_timer(x25_link);
	}
	
	
	spin_unlock_irqrestore(&x25_link->bh_lock,flags);

}

/*==============================================================
 * x25_lapb_connect_indication
 *
 * Description:
 * 	The lapb layer calls this function, when it
 * 	becomes connected.  Once the lapb layer
 * 	get connected, we can start the x25 protocol.
 */

void x25_lapb_connect_indication(void *x25_id,int reason)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;
	unsigned long flags;

	printk(KERN_INFO "%s: X25 Link connecting (tx restart req) ... \n",
			x25_link->name);

	spin_lock_irqsave(&x25_link->bh_lock,flags);

	clear_bit(BH_AWAITING_KICK,&x25_link->tq_working);
	x25_link->state = X25_ST_RESTART_REQ_SNT_R2;
	x25_clear_link_queues(x25_link,X25_ST_RESTART_REQ_SNT_R2);

	/* Some networks don't like us to send RESATART on ABM */
	if(!(x25_link->cfg.X25_protocol_options & NO_RESTART_REQ_ON_ENTER_ABM)){
		x25_send_control(x25_link,0,0,RESTART_REQUEST_PKT,0,0,NULL,0);
	}
	
	spin_unlock_irqrestore(&x25_link->bh_lock,flags);

}

/*==============================================================
 * x25_lapb_disconnect_confirm
 *
 * Description:
 *
 * 	The lapb layer calls this function, when it
 * 	becomes disconnected.  
 *
 */

void x25_lapb_disconnect_confirm(void *x25_id,int reason)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;
	unsigned long flags;
	
	printk(KERN_INFO "%s: X25 Link disconnected (lapb down)!\n",x25_link->name);

	spin_lock_irqsave(&x25_link->bh_lock,flags);
	
	x25_link->state = X25_ST_HDLC_NOT_IN_ABM;

	if (x25_link->api_sk_id){ 
	
		/* NC: Allow user listen socket to 
		 * stay connected to the driver, this way
		 * no incomint calls would be missed when
		 * the x25 layer comes back up */
		  
		wanpipe_api_listen_rx(NULL,x25_link->api_sk_id);
#if 0
		__sock_put(x25_link->api_sk_id);
		x25_link->api_sk_id=NULL;
#endif	
	}

	if (x25_link->dsp_sk_id){
		wanpipe_api_listen_rx(NULL,x25_link->dsp_sk_id);
		__sock_put(x25_link->dsp_sk_id);
		x25_link->dsp_sk_id=NULL;
	}
	
	x25_clear_link_queues(x25_link, X25_ST_HDLC_NOT_IN_ABM);
	
	spin_unlock_irqrestore(&x25_link->bh_lock,flags);
}



/*==============================================================
 * x25_lapb_disconnect_indication
 *
 * Description:
 *
 * 	The lapb layer calls this function, when it
 * 	becomes disconnected.  
 */

void x25_lapb_disconnect_indication(void *x25_id,int reason)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;
	unsigned long flags;

	printk(KERN_INFO "%s: X25 Link disconnected (lapb down)!\n",x25_link->name);

	spin_lock_irqsave(&x25_link->bh_lock,flags);

	x25_link->state = X25_ST_HDLC_NOT_IN_ABM;
	if (x25_link->api_sk_id){
		/* NC: Allow user listen socket to 
		 * stay connected to the driver, this way
		 * no incomint calls would be missed when
		 * the x25 layer comes back up */
		  
		wanpipe_api_listen_rx(NULL,x25_link->api_sk_id);
#if 0
		__sock_put(x25_link->api_sk_id);
		x25_link->api_sk_id=NULL;
#endif	
	}

	if (x25_link->dsp_sk_id){
		wanpipe_api_listen_rx(NULL,x25_link->dsp_sk_id);
		__sock_put(x25_link->dsp_sk_id);
		x25_link->dsp_sk_id=NULL;
	}

	x25_clear_link_queues(x25_link, X25_ST_HDLC_NOT_IN_ABM);

	spin_unlock_irqrestore(&x25_link->bh_lock,flags);

}


/*==============================================================
 * x25_lapb_bh_kick
 *
 * Description:
 * 	If the x25 link bh cannot transmit data to
 * 	the lapb layer it sets the BH_AWAITING_KICK
 * 	flag and waits for the lapb to kick it.
 *
 * 	The lapb layer calls this function each
 * 	time it successfuly sends data down
 * 	and frees some of its tx buffers.  
 *
 * 	If the BH_AWAITING_KICK is set trigger
 * 	the BH TX.
 */

void x25_lapb_bh_kick(void *x25_id)
{
	x25_link_t *x25_link = (x25_link_t *)x25_id;
	
	if (!test_and_clear_bit(BH_AWAITING_KICK,&x25_link->tq_working))
		return;

	if (x25_trigger_link_bh(x25_link,0)){
		set_bit(BH_AWAITING_KICK,&x25_link->tq_working);
	}
}

void x25_get_active_inactive(void *x25_id, wp_stack_stats_t *wp_stats)
{
	x25_link_t*		x25_link = (x25_link_t*)x25_id;
	struct net_device*	dev;
	unsigned long 		flags;

	WAN_ASSERT1((!x25_link));
	
	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		return;
	}

	if (x25_link->state == X25_ST_PKT_LVL_RDY_R1){
		wp_stats->x25_link_active++;
	}else{
		wp_stats->x25_link_inactive++;
	}
		
	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	for (dev=x25_link->dev_list; dev; dev = *(struct net_device **)dev->priv){

		x25_lcn_t *x25_lcn = dev_to_x25(dev);
		
		if (!x25_lcn){
			continue;
		}

		if (x25_lcn->dsp_id && IS_FUNC_CALL(x25_dsp_protocol, dsp_get_active_inactive)){
			x25_dsp_protocol.dsp_get_active_inactive(x25_lcn->dsp_id,wp_stats);
		}

		if (x25_lcn->state == X25_ST_DATA_TRANSFER_P4){
			wp_stats->x25_active++;
		}else{
			wp_stats->x25_inactive++;
		}
	}

	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);
}


int x25_get_map(void *x25_id, struct seq_file *m)
{
	x25_link_t*		x25_link = (x25_link_t*)x25_id;
	struct net_device*	dev;
	static char		prefix[100];
	char			line[500];
	unsigned long 		flags;


	if (test_bit(LINK_DOWN,&x25_link->tq_working)){
		return 0;
	}

	PROC_ADD_LINE(m,":%s:%c", x25_link->name,
				(x25_link->state == X25_ST_PKT_LVL_RDY_R1) ? '*':' ');

	WP_READ_LOCK(&x25_link->dev_list_lock,flags);
	
	for (dev=x25_link->dev_list; dev; dev = *(struct net_device **)dev->priv){

		x25_lcn_t *x25_lcn = dev_to_x25(dev);
		x25_lcn_t *x25_sw_lcn;

		if (!x25_lcn)
			continue;
		
		if (test_bit(LCN_DOWN,&x25_lcn->critical) || !(dev->flags&IFF_UP)){
			continue;
		}

		if (x25_lcn->dsp_id){
			sprintf(line, "%15s%s:%s:%i:dsp:%s:%i:%-15s\n",
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn),
				((struct net_device*)x25_lcn->dsp_id)->name,
				x25_lcn->pid,
				x25_lcn->dlabel);
	
		}else if (test_bit(LCN_CONFIG_PVC,&x25_lcn->lcn_config) && 
			  atomic_read(&x25_lcn->lcn)){

			sprintf(line, "%15s%s:%s:%i:api:pvc:%i:%-15s\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn),x25_lcn->pid,
				x25_lcn->dlabel);
			
		}else if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
			sprintf(line, "%15s%s:%s:%i:api:svc:%i:%-15s\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn),x25_lcn->pid,
				x25_lcn->dlabel);
		
		}else if ((x25_sw_lcn=(x25_lcn_t *)x25_lcn->x25_sw_lcn) != NULL){
			sprintf(line, "%15s%s:%s:%i:sw:%s\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn),
				x25_sw_lcn->dev->name);
		
		}else if (atomic_read(&x25_lcn->lcn) && x25_lcn->usedby == TCPIP) {
			sprintf(line, "%15s%s:%s:%i:tcpip:\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn));
		
		}else if (atomic_read(&x25_lcn->lcn) == 0 && x25_lcn->usedby == API){
			sprintf(line, "%15s%s:%s:%i:free:\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn));
		}else{
			sprintf(line, "%15s%s:%s:%i:ERROR:used=0x%lX:usedby=0x%X\n", 
				x25_lcn->label, prefix, dev->name,
				atomic_read(&x25_lcn->lcn),x25_lcn->used,x25_lcn->usedby);
		}

		PROC_ADD_LINE(m, line);

		//FIXME: Check for size here

		//if (x25_lcn->dsp_id && IS_FUNC_CALL(x25_dsp_protocol, dsp_get_map)){
		//	x25_dsp_protocol.dsp_get_map(NULL, x25_lcn->dsp_id);
		//}
	}

	WP_READ_UNLOCK(&x25_link->dev_list_lock,flags);

	return 0;
}

/*==============================================================
 * wanpipe_x25_init
 *
 * Description:
 * 	This function is called on module startup.
 * 	(ex: modprobe x25)
 *
 * 	Register the x25 callback functions to the 
 * 	lapb protocol. 
 *
 * 	Register the x25 callback functions to the
 * 	wanmain which are used by the socket or
 * 	upper layer.
 */

static int __init wanpipe_x25_init(void)
{
	int err;
	struct lapb_x25_register_struct x25_reg;
	struct wanpipe_x25_register_struct x25_api_reg;
	

        if (WANPIPE_VERSION_BETA){
		DEBUG_EVENT("WANPIPE X25 for Linux. Beta %s.%s %s %s\n",
			WANPIPE_VERSION, WANPIPE_SUB_VERSION, 
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
	}else{
		DEBUG_EVENT("WANPIPE X25 for Linux. Stable %s.%s %s %s\n",
			WANPIPE_VERSION, WANPIPE_SUB_VERSION, 
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
	}         
	
	x25_link_list=NULL;
	x25_link_lock=RW_LOCK_UNLOCKED;
	 
	atomic_set(&x25_skb_alloc,0);
	atomic_set(&x25_mem_alloc,0);

	UNREG_PROTOCOL_FUNC(x25_dsp_protocol);

#ifdef CONFIG_PROC_FS
	if (x25_proc_init() != X25_OK)
		return -ENOMEM;
#endif
	
	x25_reg.x25_lapb_connect_indication = x25_lapb_connect_indication;
	x25_reg.x25_setparms = x25_setparms;
	x25_reg.x25_lapb_bh_kick = x25_lapb_bh_kick;
	x25_reg.x25_receive_pkt = x25_receive_pkt;
	
	x25_reg.x25_register = x25_register;
	x25_reg.x25_unregister_link = x25_unregister_link;
	x25_reg.x25_unregister = x25_unregister;
	
	x25_reg.x25_lapb_disconnect_confirm = x25_lapb_disconnect_confirm;
	x25_reg.x25_lapb_disconnect_indication = x25_lapb_disconnect_indication;
	x25_reg.x25_lapb_connect_confirm = x25_lapb_connect_confirm; 

	x25_reg.x25_get_map = x25_get_map;
	x25_reg.x25_get_active_inactive=x25_get_active_inactive;
	
	err = register_lapb_x25_protocol(&x25_reg);
	if (err)
		return err;

	x25_api_reg.x25_bind_api_to_svc = x25_bind_api_to_svc;
	x25_api_reg.x25_bind_listen_to_link = x25_bind_listen_to_link;
	x25_api_reg.x25_unbind_listen_from_link = x25_unbind_listen_from_link;
	
	x25_api_reg.x25_dsp_register = x25_dsp_register;
	x25_api_reg.x25_dsp_unregister = x25_dsp_unregister;
	x25_api_reg.x25_dsp_setparms = x25_dsp_setparms;

	memset(&x25_link_num,0,sizeof(x25_link_num));
	
	return register_wanpipe_x25_protocol(&x25_api_reg);
}

int register_x25_dsp_protocol (struct x25_dsp_register_struct * dsp_reg)
{
	if (IS_PROTOCOL_FUNC(x25_dsp_protocol))
		return -EBUSY;
	
	memcpy(&x25_dsp_protocol, dsp_reg,sizeof(struct x25_dsp_register_struct));
	REG_PROTOCOL_FUNC(x25_dsp_protocol);
	return 0;
}

void unregister_x25_dsp_protocol (void)
{
	UNREG_PROTOCOL_FUNC(x25_dsp_protocol);
}

EXPORT_SYMBOL(register_x25_dsp_protocol);
EXPORT_SYMBOL(unregister_x25_dsp_protocol);
EXPORT_SYMBOL(x25_get_svc_dev);
EXPORT_SYMBOL(x25_get_call_info);
EXPORT_SYMBOL(x25_dsp_id_unregister);
EXPORT_SYMBOL(x25_dsp_is_svc_ready);



MODULE_AUTHOR("Nenad Corbic <ncorbic@sangoma.com>");
MODULE_DESCRIPTION("The X.25 link layer protocol");
MODULE_LICENSE("GPL");

void __exit wanpipe_x25_exit (void)
{
#if CONFIG_PROC_FS
	 x25_proc_cleanup();
#endif
	 unregister_lapb_x25_protocol();
	 unregister_wanpipe_x25_protocol();

	 printk(KERN_INFO "WANPIPE X25: Unloading module\n");

	 if (atomic_read(&x25_mem_alloc)){
		printk(KERN_INFO "X25: Warning releasing module with used memory=%i\n",
				atomic_read(&x25_mem_alloc));
	 }
	 if (atomic_read(&x25_skb_alloc)){
		printk(KERN_INFO "X25: Warning releasing module with used skb memory=%i\n",
				atomic_read(&x25_skb_alloc));
	 }
}

module_init(wanpipe_x25_init);
module_exit(wanpipe_x25_exit);



