#include <linux/wanpipe_includes.h>
#include <linux/wanpipe_defines.h>
#include <linux/wanpipe_debug.h>
#include <linux/wanpipe_common.h>
#include <linux/wanpipe_cfg.h>
#include "wanpipe_x25.h"

extern struct x25_dsp_register_struct x25_dsp_protocol;

int x25_decode_data_ptk(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame);


/*
 *	State machine for state r1, Packet Layer Ready
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_r1_machine(x25_link_t *x25, struct sk_buff *skb, struct x25_frame *frame)
{	
	WAN_ASSERT1((!x25));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	switch(frame->type) {

		/* in state r1, we should respond with a Restart Confirmation 
		 * packet (C2-G1) */
		case RESTART_REQUEST_PKT:	  

			if (Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				printk(KERN_INFO "R1: Invalid Restart Request\n");
				return;
			}
			
			/* halt all X.25 asynchronous timers which are currently active */
			//init_X25_asy_tmr_structs();
		
			send_global_api_oob_msg(x25,frame,NULL,0);
			
			/* flush the X.25 Data packet buffers for all LCNs */

			x25_sw_restart_request(x25);

			x25_set_lcn_states(x25, X25_ST_RESTART_REQ_SNT_R2);
			x25->state = X25_ST_RESTART_REQ_SNT_R2;

			/* if we are setup to automatically resond to a restart request
			 * then issue RESTART CONFIRM.  Otherwise, we proceed to R3 state
			 * and wait untill the user sends the RESTART CONFIRM */
			if (x25->cfg.X25_response_options & DISABLE_AUTO_RESTART_CONF){
				x25->state = X25_ST_RESTART_REQ_REC_R3;
			}else{
				x25_clear_link_queues(x25,X25_ST_PKT_LVL_RDY_R1);
				x25_set_lcn_states(x25,X25_ST_PKT_LVL_RDY_P1);
				x25_send_control(x25,0,0,RESTART_CONF_PKT,0,0,NULL,0);
			}
			/* X25_SW
			 * If there is LCN that used for switching, send Clear Request to
			 * remote switching device. Otherwise, do nothing.
			 */
			break;
		
		/* in state r1, we issue a Restart Request with a diagnostic code */
		/* #17 (C2-H1) */
		case RESTART_CONF_PKT:								

			/* check for a packet format error */
			if(Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				DEBUG_EVENT("%s: R1: Invalid Packet %i\n",
						x25->name,frame->len);
				return;
			}
	
			/* flush the X.25 Data packet buffers for all LCNs */
			if (x25->cfg.dte){
				x25_set_lcn_states(x25, X25_ST_RESTART_REQ_SNT_R2);
				x25->state = X25_ST_RESTART_REQ_SNT_R2;
				x25_send_control(x25,0,0,RESTART_REQUEST_PKT,0,17,NULL,0);
			}
			break;

		/* if we receive an asynchronous packet with a LCN of zero, 
		 * then we have an unassigned LCN situation */
		/* and we issue a Diagnostic packet with diagnostic code 
		 * #36 (C2-M1, C2-M2, C2-M3) */
			
		default:		

			//issue_Diagnostic_pkt((UCHAR)PKT_NOT_ALLOWED_UNASSIGNED_LCN);
			//pkts_Rx_unassigned_LCN_count ++;	
			break;
	}

	return; 
}

/*
 *	State machine for state r2, DTE Restart Request.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_r2_machine(x25_link_t *x25, struct sk_buff *skb, struct x25_frame *frame)
{

	WAN_ASSERT1((!x25));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));

	switch (frame->type){

		/* we have issued a Restart Request - the restarting procedure	
		 * has now been completed (C2-G2) */	
		case RESTART_REQUEST_PKT:		
	
			
			if(Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				//issue invlaid packet
				printk(KERN_INFO "%s: (%s:%d) R2: Received an invalid pakcet\n",
						x25->name,__FUNCTION__,__LINE__);
				return;
			}
			
			x25_stop_t10_t20_timer(x25);
	
			printk(KERN_INFO "%s: X25 Link connected (tx restart conf)!\n",
					x25->name);

			x25_sw_restart_request(x25);

			x25_clear_link_queues(x25,X25_ST_PKT_LVL_RDY_R1);
			x25_set_lcn_states(x25,X25_ST_PKT_LVL_RDY_P1);
			x25_send_control(x25,0,0,RESTART_CONF_PKT,0,0,NULL,0);
			
			if (x25->api_sk_id){ 
				wanpipe_api_listen_rx(NULL,x25->api_sk_id);
			}
			/* X25_SW
			 * If there is LCN that used for switching, send Clear Request to
			 * remote switching device. Otherwise, do nothing.
			 */
			break;

		/* we have issued a Restart Request - the restarting procedure */
		/* has now been completed (C2-H2) */
		case RESTART_CONF_PKT:

			if(Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				printk(KERN_INFO "%s: (%s:%d) R2: Received an invalid pakcet\n",
						x25->name,__FUNCTION__,__LINE__);
				return;
			}
			
			x25_stop_t10_t20_timer(x25);

			printk(KERN_INFO "%s: X25 Link connected (rx restart conf)!\n",
					x25->name);
		
			x25_clear_link_queues(x25,X25_ST_PKT_LVL_RDY_R1);
			x25_set_lcn_states(x25,X25_ST_PKT_LVL_RDY_P1);
			
			if (x25->api_sk_id){ 
				wanpipe_api_listen_rx(NULL,x25->api_sk_id);
			}
			break;
	
		/* these asynchronous packet types are valid only for SVCs */

		default:
		
			printk(KERN_INFO "%s: R2: Received an invalid frame %lx\n",
					x25->name,frame->type);

			//FIXME
			//issue_Diagnostic_pkt((UCHAR)PKT_NOT_ALLOWED_UNASSIGNED_LCN);
			//pkts_Rx_unassigned_LCN_count ++;	

			/* discard the call setup packet if we are in the state */
			/* r2 (C2-N2) or in state d2 (C4-F2) */
			return;
	}

	return;
}

/*
 *	State machine for state r3, DTE Restart Indication.
 *	The handling of the timer(s) is in file x25_timer.c
 */
static void x25_state_r3_machine(x25_link_t *x25, struct sk_buff *skb, struct x25_frame *frame)
{
	unsigned char diag=0;

	WAN_ASSERT1((!x25));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
		
	/* if we have previously sent a Restart Request packet, 
	 * then all call setup/clearing, Data, Interrupt, 
	 * Flow Control and Reset packets are discarded (C2-N2) */

        /* The only way we get out of R3 state is buy sending
	 * a REQUEST CONFIRMATION.
	 *
	 * If we receive an REQUEST CONFIRMATION in the mean time
	 * we send a RESTART REQUEST as an error condition and
	 * go back to R2 */
	
	switch(frame->type) {
	
		/* we have already received a Restart Indication, so we discard */
		/* the packet (C2-G3) */
		case RESTART_REQUEST_PKT:
			printk(KERN_INFO "R3: Got Restart Reqest: discarding\n");

			if (Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				DEBUG_EVENT("%s: R2: Invalid Packet %i\n",
						x25->name,frame->len);
				return;
			}

			return;

		/* we have previously received a Restart Indication, so we issue */
		/* a Restart Request with a diagnostic code #19/#18 (C2-H3) */	
		case RESTART_CONF_PKT:			
			
			printk(KERN_INFO "R3: Got Restart Conf: discarding\n");
			

			if(Restrt_Registr_Diag_pkt_error(x25,skb,frame)){	
				printk(KERN_INFO "R2: Invalid Pakcet\n");
				return;
			}
		
			if (x25->cfg.dte){
				diag=PKT_TYPE_INVALID_STATE_r3;
			}else{
				diag=PKT_TYPE_INVALID_STATE_r2;
			}

			x25_sw_restart_request(x25);

			x25->state = X25_ST_RESTART_REQ_SNT_R2;
			x25_set_lcn_states(x25, X25_ST_RESTART_REQ_SNT_R2);
			x25_send_control(x25,0,0,RESTART_REQUEST_PKT,0,diag,NULL,0);
			/* X25_SW
			 * If there is LCN that used for switching, send Clear Request to
			 * remote switching device. Otherwise, do nothing.
			 */
			return;

		default:
			/* we have received a Restart Request, 
			 * so issue a Restart Request with a diagnostic 
			 * code #19 to indicate the error (C2-N3)*/

			printk(KERN_INFO "R3: Got invalid state\n");

			x25_sw_restart_request(x25);

			x25->state = X25_ST_RESTART_REQ_SNT_R2;
			x25_set_lcn_states(x25, X25_ST_RESTART_REQ_SNT_R2);
			x25_send_control(x25,0,0,RESTART_REQUEST_PKT,0,
					 x25->cfg.dte ? 
					   PKT_TYPE_INVALID_STATE_r3 :
				           PKT_TYPE_INVALID_STATE_r2,
					 NULL,0);
			/* X25_SW
			 * If there is LCN that used for switching, send Clear Request to
			 * remote switching device. Otherwise, do nothing.
			 */
			return;	
	}

	/* if we receive an asynchronous packet with a LCN of zero, 
	 * then we have an unassigned LCN situation */
	/* and we issue a Diagnostic packet with diagnostic code 
	 * #36 (C2-M1, C2-M2, C2-M3) */

	return;
}


void x25_state_p1_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{

	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));


	/* process a packet when in the 'packet level ready' state */

	switch(frame->type) {
	
	case CALL_REQUEST_PKT:						
		/* handle an incoming Call Request packet (C3-D1) */
		lcn_state_change(x25_lcn,X25_ST_CALL_REQ_REC_P3);
		//NOTE: This is handed before LCN state machines
		//handle_Rx_Call_Request_pkt(x25_lcn,skb,frame);	
		break;


	case CLEAR_REQUEST_PKT:
		/* handle an incoming Clear Request packet (C3-F1) */
		handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
		break;

	default:
		/* all other packets received constitute an error condition, so 
		 * we issue a Clear Request with a diagnostic code #20 (C3-E1, 
		 * C3-G1, C3-H1) */
		//issue_Clear_Request((UCHAR)PKT_TYPE_INVALID_STATE_p1);
		lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
		x25_send_control(x25_lcn->x25_link, 
				 0, 
				 atomic_read(&x25_lcn->lcn),
				 CLEAR_REQUEST_PKT,
				 0,
				 PKT_TYPE_INVALID_STATE_p1,NULL,0);
		break;
	}
}



void x25_state_p2_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	int err;
	/* we have previously sent a Call Request packet */

	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));

	
	switch(frame->type) {

		case CALL_REQUEST_PKT:
			
			if(x25_lcn->x25_link->cfg.dte) {
				
				/* call collision state - for a DTE, we ignore 
				 * the incoming call (C3-D2) increment the number 
				 * of Call Request packets received on this LCN */
				//ptr_X25_Unnum_pkt_stats->Call_collision_count ++;
				lcn_state_change(x25_lcn,X25_ST_CALL_COLLISION_P5);
			}else{
				/* handle an incoming Call Request packet */
				lcn_state_change(x25_lcn,X25_ST_CALL_REQ_REC_P3);
				//ERROR COLLISSION !!!
				//handle_Rx_Call_Request_pkt(x25_lcn->x25_link,skb,frame);
			}
			break;

		case CALL_ACCEPT_PKT:
	
			/* handle an incoming Call Accept packet (C3-E2) */
			if ((err=handle_Rx_Call_Accept_pkt(x25_lcn,skb,frame)) != X25_OK){

				if (err<0){
					err = 64;
				}

				printk(KERN_INFO "Invalid Accept Packet!\n");
				lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
				x25_send_control(x25_lcn->x25_link, 
						 0, 
						 frame->lcn,
						 CLEAR_REQUEST_PKT,
						 0x13,err,NULL,0);
			}else{
				lcn_state_change(x25_lcn,X25_ST_DATA_TRANSFER_P4);
				x25_stop_t11_t21_timer(x25_lcn);
			}
			break;

		case CLEAR_REQUEST_PKT:
			/* handle an incoming Clear Request packet (C3-F2) */
			x25_stop_t11_t21_timer(x25_lcn);
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		default:
			/* all other packets received constitute an error condition, so 
			 * we issue a Clear Request with a diagnostic code #21/#22 
			 * (C3-G2, C3-H2) */
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 
				 	 0, 
				 	 atomic_read(&x25_lcn->lcn),
				 	 CLEAR_REQUEST_PKT,
				 	 0,
				 	 (x25_lcn->x25_link->cfg.dte) ? 
					 	PKT_TYPE_INVALID_STATE_p2: 
						PKT_TYPE_INVALID_STATE_p3,NULL,0);
			break;
		}
	
}


/*
 *	State machine for state p3, DTE Incoming Call.
 *	The handling of the timer(s) is in file x25_timer.c
 */
static void x25_state_p3_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	switch(frame->type) {
		
		/* handle an incoming Clear Request packet (C3-F3) */
		case CLEAR_REQUEST_PKT:
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		default:
			/* all other packets received constitute an error condition, so
			 * we issue a Clear Request with a diagnostic code #22/#21
			 * (C3-D3, C3-E3, C3-G3, C3-H3) */
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 
				 	 0, 
				 	 atomic_read(&x25_lcn->lcn),
				 	 CLEAR_REQUEST_PKT,
				 	 0,
				 	 (x25_lcn->x25_link->cfg.dte) ? 
					 	PKT_TYPE_INVALID_STATE_p3: 
						PKT_TYPE_INVALID_STATE_p2,NULL,0);

			

			break;
	}
}

/*
 *	State machine for state p4, Data Transfer.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static int x25_state_p4_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	unsigned short X25_Data_lgth = skb->len;
	int err = -EINVAL;

	WAN_ASSERT2((!x25_lcn),-EINVAL);
	WAN_ASSERT2((!skb),-EINVAL);
	WAN_ASSERT2((!frame),-EINVAL);
	
	/* process the incoming frame */
	switch(frame->type) {				
	
	case DATA_PKT:	/* a Data packet */

		/* if there is no storage space available for this 
		 * X.25 Data packet or if we are not able to queue and 
		 * RR packet for later transmission, then don't free 
		 * up the HDLC I-frame buffer - we will attempt to 
		 * process this I-frame later */
		
		if (x25_lcn->usedby == API){
			if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){

				if (IS_PVC(x25_lcn) && x25_lcn->state != X25_ST_DATA_TRANSFER_P4){
					/* It is legal for PVC to be in disconnected
					 * state.  In this state, pass the data packet
					 * through x25 state machine, and drop it in
					 * the upper layers */
					err = 0;
				}else{
					err = wanpipe_api_buf_check(x25_lcn->sk_id,
								    (skb->truesize-LGTH_X25_PACKET_HEADER));
					if (err == -EINVAL){
						printk(KERN_INFO "%s:Major Error in Socket Above: CHK BUF!!!\n",
							x25_lcn->dev->name);
						__sock_put(x25_lcn->sk_id);
						clear_bit(LCN_SK_ID,&x25_lcn->used);
						x25_lcn->sk_id = NULL;
						err=1;
					}
				}

			}else if (x25_lcn->dsp_id){
				err=-EINVAL;
				if (IS_FUNC_CALL(x25_dsp_protocol, dsp_rx_buf_check)){
					err = x25_dsp_protocol.dsp_rx_buf_check(x25_lcn->dsp_id, (skb->truesize-LGTH_X25_PACKET_HEADER));
				}
			}else if (x25_lcn->x25_sw_lcn){
				/* X25_SW
				 * If this LCN used for switching, theoretically(!) always should exists
				 * place in write_queue queue for the new received packet.
				 */
				err = X25_OK;
			}else{
				err = -EINVAL;
				if (!IS_PVC(x25_lcn)){
					printk(KERN_INFO 
						"%s: Rx packed on invalid lcn: lcn is free!\n",
						x25_lcn->dev->name);
					break;
				}else{
					if (x25_lcn->x25_link->cfg.X25_response_options &
						RESET_PVC_NET_DOWN_ON_RX_DATA){

						printk(KERN_INFO "%s: %s: Error: Rx data on unused pvc (lcn=%i)!\n",
							x25_lcn->x25_link->name,x25_lcn->dev->name,frame->lcn);

						x25_initialize_lcn(x25_lcn,
								   X25_ST_PKT_LVL_RDY_D1,
								   SOFT_INIT);

						x25_send_control(x25_lcn->x25_link,
								 0,frame->lcn,
					 			 RESET_REQUEST_PKT,
					 			 RESET_NETWORK_OUT_OF_ORDER,
								 0,
					 			 NULL,0);
						err=-EINVAL;
						break;
					}
					err=0;
					/* If this interface is PVC pass
					 * the rx packet down the state
					 * machine, and the packet will
					 * be dropped in the upper layer.
					 *
					 * It is legal for the PVC not to
					 * have an attached interface */
				}
			}

			if (err != X25_OK){
				if (net_ratelimit()){
				printk(KERN_INFO "%s: Error: Upper layer busy queuing packet!\n",
						x25_lcn->dev->name);
				}
				err=1;
				break;
			}
		}
		
		err=-EINVAL;
		
		if (x25_decode_data_ptk(x25_lcn, skb, frame) != X25_OK){
			break;
		}
	
		/* The decode will strip the x25 header thus, update
		 * the data length */
		X25_Data_lgth = skb->len;
		
		/* calculate the length of the actual Data field and 
		 * check for a Data packet of excessive length */
		/* (C6 - Note 4) - note that we store the current 
		 * Rx packet size for use in the M-bit check below */

		if(X25_Data_lgth > x25_lcn->rx_pkt_size){
		
			/* the Data packet is too long, so issue 
			 * a Reset Request with a diagnostic code #39 */
			
			printk(KERN_INFO "%s: Error: Rx packet greater than MRU (%i>%i)!\n",
					x25_lcn->dev->name,X25_Data_lgth,x25_lcn->rx_pkt_size);

			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,
					 0,frame->lcn,
					 RESET_REQUEST_PKT,
					 0,PKT_TOO_LONG,
					 NULL,0);
			break;
		}

		/* check the packet for a valid P(R) count */
		if (x25_validate_Pr(x25_lcn, frame) == X25_OK){

			/* check the validity of the P(S) field in the Data packet */ 
			
			if (x25_lcn->Pr_count != frame->Ps_count){
				
				printk(KERN_INFO "%s: Rx Data Pkt Error: invalid PS count: Orig=%i != New=%i\n",
						x25_lcn->dev->name,x25_lcn->Pr_count,frame->Ps_count);
				/* the P(S) count is invalid, so issue a Reset 
				 * Request with a diagnostic code #1 (C6-H1) */


				lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
				x25_send_control(x25_lcn->x25_link,
					 0,frame->lcn,
					 RESET_REQUEST_PKT,
					 0,INVALID_Ps_COUNT,
					 NULL,0);
				
				break;
			}

			/* Update the local Pr count */
			x25_lcn->Pr_count = (x25_lcn->Pr_count+1) % 
				((x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET) ? 
			 			MOD8_MODULUS : MOD128_MODULUS);
	
			/* check the validity of a Data packet with the D-bit set */
			if(frame->gfi & GFI_WITH_D_BIT_SET) {
		
				//??????????????????
				printk(KERN_INFO "%s: Rx Data Pkt: D-BIT Set\n",
						x25_lcn->dev->name);
				
				/* if D-bit pragmatics are not supported or D-bit 
				 * procedures were not established during 
				 * call setup for this SVC, then we issue a 
				 * Reset Request with a diagnostic code #166 or
				 * #40 (C6-I1) */
				if ((x25_lcn->x25_link->cfg.X25_protocol_options & 
							D_BIT_PRAGMATICS_NOT_SUPPORTED)){
				
					printk(KERN_INFO "%s: Rx Data Pkt Error: invalid pkt with DBIT set\n",
							x25_lcn->dev->name);

					lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
					x25_send_control(x25_lcn->x25_link,
					 	0,frame->lcn,
					 	RESET_REQUEST_PKT,
					 	0,x25_lcn->x25_link->cfg.dte ? 
							D_BIT_PROCEDURE_NOT_SUPPORTED:
							INVALID_GFI,
					 	NULL,0);

					break;
				}else{
					//?????????????????
					printk(KERN_INFO "%s: Sending out DBIT Response: RR PKT\n",
							x25_lcn->dev->name);
					x25_send_control(x25_lcn->x25_link,
						0,
						frame->lcn,
						RR_PKT,
						x25_lcn->Pr_count,0,
						NULL,0);
				}
			}

			/* if the Data packet has the M-bit set and D-bit reset 
			 * and the user Data field is partially full, 
			 * then we issue a Reset Request with a diagnostic 
			 * code #165 (C6-I1) */
			else if ((frame->mbit) && (X25_Data_lgth < x25_lcn->rx_pkt_size)) {

				printk(KERN_INFO "%s: Rx Data Error: Invalid Packet with Mbit set(%i<%i)\n",
						x25_lcn->dev->name, X25_Data_lgth, x25_lcn->rx_pkt_size);
				lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
				x25_send_control(x25_lcn->x25_link,
						gfi_decode(x25_lcn->config),
						frame->lcn,
						RESET_REQUEST_PKT,
						0,INVALID_PART_FULL_DATA_PKT,
						NULL,0);
				break;
			}

			if (x25_lcn->x25_sw_lcn == NULL) {
				/* the Data packet has a valid P(R), a valid P(S) and a user data */
				/* field with a proper format, so store this packet (C6-J1) */

				if (x25_data_tx_up (x25_lcn,skb,frame) == X25_OK){
					err = 0;
				}else{
					printk(KERN_INFO "%s: Critical ERROR!! DATA LOST !\n",
							x25_lcn->dev->name);
					break;
				}
			}else{
				/* X25_SW
				 * This LCN used for switching. Theoretically(!) always should exists
				 * place in write_queue queue for the new received packet.
				 */
				err = x25_sw_data_rx(x25_lcn, skb, frame);
			}
			
			/* We issue an RR in response to the Data packet 
			 * under the following circumstatnces:
			 * 1) the packet must not have the D-bit set AND
			 * 2) we must have received a complete window 
			 *    worth of Data packets on this LCN OR
			 * 3) we are configured to respond to all buffered 
			 *    Data packets with an RR
			 */

			/* X25_SW
			 * Send RR to remote device only for non-switching LCN.
			 */
			if ((x25_lcn->x25_sw_lcn == NULL) && 
			    (!(frame->gfi & GFI_WITH_D_BIT_SET)) && 
			    (((x25_lcn->Pr_count%x25_lcn->rx_window_size)==0) ||		
			     (x25_lcn->x25_link->cfg.X25_response_options & 
			      			ALL_DATA_PKTS_ACKED_WITH_RR))){

				if ((x25_lcn->Pr_count%x25_lcn->rx_window_size)!=0 &&
				     x25_lcn->x25_link->cfg.X25_response_options & 
				     	ALL_DATA_PKTS_ACKED_WITH_RR && 
					skb_queue_len(&x25_lcn->write_queue)){
						
						/* NC OPTIMIZATION:
						 * There is an I-Frame pending for
						 * TX thus, skip sending the RR
						 * packet. */
				}else{
				
					x25_send_control(x25_lcn->x25_link,
							gfi_decode(x25_lcn->config),
							frame->lcn,
							RR_PKT,
							x25_lcn->Pr_count,0,
							NULL,0);
				}
			}

		}else{
			err=-EINVAL;
			DEBUG_EVENT("%s:%s: Error Rx DATA with invalid Pr cnt\n",
					x25_lcn->x25_link->name,x25_lcn->dev->name);

			/* the P(S) count is invalid, so issue a 
			 * Reset Request with a diagnostic code #1 (C6-H1) */
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,
					0,
					frame->lcn,
					RESET_REQUEST_PKT,
					0, INVALID_Ps_COUNT,
					NULL, 0);
		}

			
		break;

	case RR_PKT:
		/* an RR or RNR frame */
		/* increment the number of RR packets received on this LCN */
		//ptr_X25_Data_Sup_pkt_stats->RR_pkts_Rx_count ++;
		
	case RNR_PKT:
	
		err=-EINVAL;

		/* we have already checked for the packet being too short, 
		 * so only check for a too long packet */
		if (x25_decode_data_ptk(x25_lcn, skb, frame) != X25_OK){
			break;
		}

#if BH_DEBUG > 2
		if (frame->type != RR_PKT){
			printk(KERN_INFO "GOT RNR PACKET\n");
		}else{
			printk(KERN_INFO "GOT RR PACKET Pr=%i\n", frame->Pr_count);
		}
#endif

			/* process a packet length error */
		if (packet_length_error(x25_lcn->x25_link, frame)){
			DEBUG_EVENT("%s: Invalid RR/RNR Packet Length %i\n!\n",
					x25_lcn->dev->name,frame->len);
			/* FIXME issue a reset or something */
			break;		
		}
	
		/* check the P(R) count in the packet */
		/* process an RR or RNR with a valid P(R) (C6-D1, C6-D2) - 
		 * if we received an RNR, then we set our */
		/* internal 'RNR received' flag */

		/* check the packet for a valid P(R) count */
		if (x25_validate_Pr(x25_lcn, frame) == X25_OK){

			set_LCN_misc_status_bits(x25_lcn,(frame->type == RNR_PKT) ? 
					(UCHAR)RNR_RX_ON_LCN : (UCHAR)~RNR_RX_ON_LCN);
		
			/* if we received an RNR packet, then indicate that the Tx window */
			/* is closed */
			if(frame->type == RNR_PKT){	
				//printk(KERN_INFO "RECEIVED RNR: WINDOW CLOSE \n");
				set_bit(TX_WINDOW_CLOSED,&x25_lcn->critical);
			}else{
				//printk(KERN_INFO "RECEIVED RR: WINDOW OPEN \n");
				clear_bit(TX_WINDOW_CLOSED,&x25_lcn->critical);
			}

			if (x25_lcn->x25_sw_lcn){
				/* X25_SW
			 	 * This LCN used for switching. 
				 * Resend RR/RNR to remote switching device.
				 */
//				printk(KERN_INFO "%s: Debug X25 Event: Rx RR :Lcn=%d:Pr=0x%02X\n",
//					x25_lcn->dev->name,
//					atomic_read(&x25_lcn->lcn), frame->Pr_count);
				x25_sw_send_control(x25_lcn->x25_sw_lcn, skb?skb:NULL, frame);
			}
			
#if BH_DEBUG == 2
			printk(KERN_INFO "RR WAKING QUEUE Crit %x\n",x25_lcn->critical);
			netif_wake_queue(x25_lcn->dev);
#endif

		}else{
			printk(KERN_INFO "%s:%s: Error Rx %s with invalid Pr cnt State=%i\n",
					x25_lcn->x25_link->name,x25_lcn->dev->name,
					(frame->type == RNR_PKT) ? "RNR" : "RR",x25_lcn->state);

			/* the P(S) count is invalid, so issue a 
			 * Reset Request with a diagnostic code #1 (C6-H1) */
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,
					0,
					frame->lcn,
					RESET_REQUEST_PKT,
					0, INVALID_Pr_COUNT,
					NULL, 0);
		}
		break;


	default:
		switch(frame->type) {	


		case CALL_REQUEST_PKT:
		case CALL_ACCEPT_PKT:
		case CLEAR_CONF_PKT:
			printk(KERN_INFO "%s: X25 Event: Error Rx %s on connected LCN=%i !\n",
						x25_lcn->dev->name, 
						(frame->type == CALL_REQUEST_PKT) ? "Call Request":
						(frame->type == CALL_ACCEPT_PKT) ? "Call Accept" : "Clear Conf",
						atomic_read(&x25_lcn->lcn));
			
			/* we issue a Clear Request with a diagnostic 
			 * code #23 (C3-D4, C3-E4, C3-G4) */
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link,
					 0,
					 frame->lcn,
					 CLEAR_REQUEST_PKT,
					 0,
					 PKT_TYPE_INVALID_STATE_p4,NULL,0);
			break;

			
		/* handle a valid Clear Request packet (C3-F4) */
		case CLEAR_REQUEST_PKT:  			
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		
			/* handle a valid Reset Request packet (C4-D1) */
		case RESET_REQUEST_PKT:				
			handle_Rx_Reset_Request_pkt(x25_lcn,skb,frame);
			break;
			
		/* handle a valid Reset Confirmation packet */
		case RESET_CONF_PKT:				

			if (IS_PVC(x25_lcn) && x25_lcn->usedby == API){
				/* Unused pvc can send a reset with network down,
				 * in case date packets arrive and pvc is not
				 * used by an API. */
				
				if (!test_bit(LCN_SK_ID,&x25_lcn->used) || !x25_lcn->sk_id){
					x25_initialize_lcn(x25_lcn,
							   X25_ST_PKT_LVL_RDY_D1,
							   SOFT_INIT);
					break;
				}
			}
				/* we issue a Reset Request with a diagnostic code #27 (C4-E1) */
			printk(KERN_INFO "%s: X25 Event: Error Rx Reset Conf on connected LCN=%i !\n",
					x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));

			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,0,
					 frame->lcn,
					 RESET_REQUEST_PKT,0,
					 PKT_TYPE_INVALID_STATE_d1,NULL,0);
			break;

		case INTERRUPT_PKT:
		
			/* Respond right away */
			printk(KERN_INFO "%s: X25 Event: Rx Interrupt Request : LCN=%i !\n",
					x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));

			
			printk(KERN_INFO "%s: X25 Event: Tx Interrupt Conf    : LCN=%i\n",
					x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));
			
			lcn_state_change(x25_lcn,X25_ST_PKT_LVL_RDY_I1);
			x25_send_control(x25_lcn->x25_link,0,
					 frame->lcn,
					 INTERRUPT_CONF_PKT,0,
					 0,NULL,0);
			break;

		case INTERRUPT_CONF_PKT:
			
			/* if this Interrupt is not received in response 
			 * to an Interrupt packet transmitted, we */
			/* issue a Reset Request with a diagnostic code #43 (C5_A1) */
			
			printk(KERN_INFO "%s: X25 Event: Error Rx Interrupt Conf on connected LCN=%i !\n",
					x25_lcn->dev->name,atomic_read(&x25_lcn->lcn));

			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,0,
					 frame->lcn,
					 RESET_REQUEST_PKT,0,UNAUTH_INTERRUPT_CONF,NULL,0);

			break;

		default:
			break;
		}
		
		break;
	}
	
	return err;
}



/*
 *	State machine for state p4, Clear Request Sent.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_p6_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{

	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));

	
	switch(frame->type) {
	
		case CLEAR_REQUEST_PKT: 
			/* handle an incoming Clear Request packet (C3-F6) */
			x25_stop_t13_t23_timer(x25_lcn);
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		case CLEAR_CONF_PKT:
			/* attempt to buffer the incoming Clear Confirmation (C3-G6) */
			CALL_LOG(x25_lcn->x25_link,"%s: X25 Event: Rx Clear Conf    :Lcn=%i\n",
					x25_lcn->dev->name,frame->lcn);
			x25_log_proc_event(x25_lcn->x25_link,x25_lcn->dev->name,"Rx Clear Conf",
			   		   frame->lcn,frame->cause,frame->diag,NULL);

			
			if (!x25_lcn->x25_sw_lcn &&
			    x25_lcn->usedby == API){ 
				send_api_oob_msg(x25_lcn,
			     			 frame,
						 NULL,
						 0);
			}
	
			{
			x25_lcn_t* x25_sw_lcn = (x25_lcn_t*)x25_lcn->x25_sw_lcn;
			if (x25_sw_lcn){
				unsigned long flags;
				/* X25_SW
				 * This LCN used for switching.
				 * 1. Send Clear Confirmation to remote switching device.
				 * 2. Initialize switching LCN.
				 * 3. Set `x25_sw_lcn` to NULL for both switching LCN's.
				 */ 

				x25_sw_send_control(x25_sw_lcn, NULL, frame);
				spin_lock_irqsave(&x25_sw_lcn->x25_link->bh_lock,flags);
				x25_initialize_lcn(x25_sw_lcn,X25_ST_PKT_LVL_RDY_P1,HARD_INIT);
				x25_initialize_lcn(x25_lcn,X25_ST_PKT_LVL_RDY_P1,HARD_INIT);
				spin_unlock_irqrestore(&x25_sw_lcn->x25_link->bh_lock,flags);
			}else{
				x25_initialize_lcn(x25_lcn,X25_ST_PKT_LVL_RDY_P1,HARD_INIT);
			}
			}
			break;

		default:	
			/* all other valid packets are discarded (C3-D6, C3-E6, C3-H6) */
			break;
	}
}


/*
 *	State machine for state p7, Clear Request Received.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_p7_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{

	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	//printk(KERN_INFO "P7 : GOT PACKET %x\n",frame->type);

	/* we have previously received a Clear Request packet */
	switch(frame->type) {
	
		case CLEAR_REQUEST_PKT:
			/* discard an additional Clear Request packet (C3-F7) */
			break;

		default:		
			/* all other packets received constitute an 
			 * error condition, so we issue a Clear Request 
			 * with a diagnostic code #26/#25 
			 * (C3-D7, C3-E7, C3-G7, C3-H7) */
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 0, 
					frame->lcn, CLEAR_REQUEST_PKT, 0,
					(x25_lcn->x25_link->cfg.dte) ? 
					   	PKT_TYPE_INVALID_STATE_p7 :
						PKT_TYPE_INVALID_STATE_p6,NULL,0);
			break;
	}
}


/*
 *	State machine for state d2, Reset Request Sent.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_d2_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	//printk(KERN_INFO "D2 : GOT PACKET %x\n",frame->type);

	/* we have previously received a Reset Request packet */
	switch(frame->type) {
	
		case RESET_REQUEST_PKT:
			handle_Rx_Reset_Request_pkt(x25_lcn,skb,frame);
			break;
			
		case RESET_CONF_PKT:
			printk(KERN_INFO "%s: X25 Event: Rx Reset Conf   :Lcn=%i\n",
					x25_lcn->dev->name,frame->lcn);

			if (x25_lcn->usedby == API){ 
				if (test_bit(LCN_SK_ID,&x25_lcn->used) && x25_lcn->sk_id){
					send_api_oob_msg(x25_lcn,
			     				 frame,
							 NULL,
						 	0);
				}else if (x25_lcn->dsp_id){
					if (IS_FUNC_CALL(x25_dsp_protocol, dsp_svc_reset)){
						x25_dsp_protocol.dsp_svc_reset(x25_lcn->dsp_id);	
					}
				}
			}

			x25_initialize_lcn(x25_lcn,X25_ST_PKT_LVL_RDY_D1,SOFT_INIT);
			break;

		/* handle a valid Clear Request packet (C3-F4) */
		case CLEAR_REQUEST_PKT:  			

			if (IS_PVC(x25_lcn)){
				/* discard clear call on PVC */
				return;
			}
				
			x25_stop_t12_t22_timer(x25_lcn);
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		default:
			/*discard all other packets */
			break;
	}
	return;
}




/*
 *	State machine for state d3, Reset Request Received.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_d3_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	//printk(KERN_INFO "D3 : GOT PACKET %x\n",frame->type);

	/* we have previously received a Reset Request packet */
	switch(frame->type) {
	
		case RESET_REQUEST_PKT:
			break;

		/* handle a valid Clear Request packet (C3-F4) */
		case CLEAR_REQUEST_PKT:  			

			if (IS_PVC(x25_lcn)){
				/* If we receive Clear call on PVC, drop down
				 * to discard state */
				goto state_d3_discard;
			}
			
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;
	
		default:
state_d3_discard:
			/*discard all other packets */
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,0,
				 frame->lcn,
				 RESET_REQUEST_PKT,0,
				 PKT_TYPE_INVALID_STATE_d3,NULL,0);
			break;
	}
	return;
}

/*
 *	State machine for state i2, Interrupt Request Sent.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_i2_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{

	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	//printk(KERN_INFO "I2 : GOT PACKET %x\n",frame->type);

	/* we have previously received a Reset Request packet */
	switch(frame->type) {
	
		case INTERRUPT_CONF_PKT:
			lcn_state_change(x25_lcn,X25_ST_PKT_LVL_RDY_I1);
			break;
	
		/* handle a valid Clear Request packet (C3-F4) */
		case CLEAR_REQUEST_PKT:  			
			if (IS_PVC(x25_lcn)){
				goto state_i2_drop;
			}
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

		/* handle a valid Reset Request packet (C4-D1) */
		case RESET_REQUEST_PKT:				
			handle_Rx_Reset_Request_pkt(x25_lcn,skb,frame);
			break;

		default:
state_i2_drop:
			/* if this Interrupt is not received in response 
		 	 * to an Interrupt packet transmitted, we issue a 
			 * Reset Request with a diagnostic code #43 (C5_A1) */
			
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,0,
				 frame->lcn,
				 RESET_REQUEST_PKT,0,
				 UNAUTH_INTERRUPT_CONF,NULL,0);
			break;
	}
	return;
}

/*
 *	State machine for state i3, Interrupt Request Received.
 *	The handling of the timer(s) is in file x25_timer.c.
 */
static void x25_state_i3_machine(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT1((!x25_lcn));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	//printk(KERN_INFO "I3 : GOT PACKET %x\n",frame->type);

	/* we have previously received a Interrupt Request packet
	 * until we respond to the interrupt packet issue an error
	 * for each packet */
	switch(frame->type) {
	
		case INTERRUPT_PKT:
			lcn_state_change(x25_lcn,X25_ST_PKT_LVL_RDY_I1);
			break;
	
		/* handle a valid Clear Request packet (C3-F4) */
		case CLEAR_REQUEST_PKT:  			
			if (IS_PVC(x25_lcn)){
				goto state_i3_drop;
			}
			handle_Rx_Clear_Request_pkt(x25_lcn,skb,frame);
			break;

	
		/* handle a valid Reset Request packet (C4-D1) */
		case RESET_REQUEST_PKT:				
			handle_Rx_Reset_Request_pkt(x25_lcn,skb,frame);
			break;
			
		default:
state_i3_drop:
			/* if this Interrupt is not received in response 
		 	 * to an Interrupt packet transmitted, we issue a 
			 * Reset Request with a diagnostic code #43 (C5_A1) */
			
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
			x25_send_control(x25_lcn->x25_link,0,
				 frame->lcn,
				 RESET_REQUEST_PKT,0,
				 UNAUTH_INTERRUPT_CONF,NULL,0);
			break;
	}
	return;
}


static void x25_state_no_dev_machine(x25_link_t *x25_link, struct sk_buff *skb, struct x25_frame *frame)
{

	int err=0;
	int cause=0;
	struct net_device *dev=NULL;
	x25_lcn_t *x25_lcn=NULL;
	
	WAN_ASSERT1((!x25_link));
	WAN_ASSERT1((!skb));
	WAN_ASSERT1((!frame));
	
	switch (frame->type){
		
	case CALL_REQUEST_PKT:
		err = handle_Rx_Call_Request_pkt (x25_link, skb, frame);
		if (err != X25_OK){

			if (err<0){
				cause=0x13;
				err = 64;
			}
			
			/* Call was not accepted, if LCN was allocated
			 * set it in P6 state so incoming clear conf
			 * would initialize the channel */
			dev = x25_get_lcn_dev(x25_link,frame->lcn);	
			if (dev && dev->priv){
				x25_lcn=(x25_lcn_t*)dev->priv;
				lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			}
			x25_send_control(x25_link, 
				         0, 
					 frame->lcn,
					 CLEAR_REQUEST_PKT,
					 cause,err,NULL,0);
			if (dev) {
				X25_PUT(dev);
			}
		}
		break;

	case CLEAR_REQUEST_PKT:
		x25_send_control(x25_link, 
			         0, 
				 frame->lcn,
				 CLEAR_CONF_PKT,
				 0,32,NULL,0);
		break;

	case RESET_REQUEST_PKT:
		x25_send_control(x25_link, 
			         0, 
				 frame->lcn,
				 RESET_CONF_PKT,
				 0,32,NULL,0);
		break;

	default:
		//printk(KERN_INFO "RX Packet 0x%02X, No defined LCN\n",
		//		frame->type);
		break;
	}
}




/*
 *	Actions taken by the DTE upon receipt of packets in any
 *	state of the packet layer DTE/DXE interface as preceived by 
 *	the DTE.
 *	The handling of the timer(s) is in file x25_timer.c.
 *
 *	TABLE 31 Packety Layer State Table - Any State (pg 105)
 */
static int x25_pkt_check_any(x25_link_t *x25_link, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT2((!x25_link),-EINVAL);
	WAN_ASSERT2((!skb),-EINVAL);
	WAN_ASSERT2((!frame),-EINVAL);
	
	frame->len = skb->len;
	
	/* Any packet less than 2 octets */
	if (skb->len < 2){
		/* A=DIAG D=28 */
		return 28;
	}

#if 0
	{
		int i;
		printk(KERN_INFO "\nRX PACKET: "); 
		for (i=0;i<skb->len;i++){
			printk(KERN_INFO " %x",skb->data[i]);
		}
		printk(KERN_INFO "\n");
	}
#endif
	
	frame->gfi = (skb->data[0] >> 4) & 0x0f;

	frame->lcn = (skb->data[0]&0x0f << 8) | skb->data[1];

	frame->type = skb->data[2];

	//printk(KERN_INFO "RX GOT PACKET : GFI %x, LCN %i\n",frame->gfi,frame->lcn);
	
	/* Invalid general format identifier: A=DIAG D=40 */
	if (!(frame->gfi & GFI_FOR_MOD_128) &&  !(frame->gfi & GFI_FOR_MOD_8)){
		/* A=DIAG D=40   */
		DEBUG_EVENT("%s: Error: Rx Invalid GFI 0x%X pkt\n",
					x25_link->name,frame->gfi);
		return 40;
	}

	/* if the Q, D or MOD 128 bits of the GFI are set, we must check */
	if (frame->gfi & GFI_MOD_128_BIT_SET) {

		if (!(x25_link->cfg.mode & GFI_MOD_128_BIT_SET)){
			DEBUG_EVENT("%s: Rx Invalid MOD128 pkt. Local cfg MOD8!\n",
					x25_link->name);
			return 40;
		}	
		

	}else if (frame->gfi & GFI_MOD_8_BIT_SET){

		if (!(x25_link->cfg.mode & GFI_MOD_8_BIT_SET)){
			DEBUG_EVENT("%s: Rx Invalid MOD8 pkt. Local cfg MOD128!\n",
					x25_link->name);
			return 40;
		}		

	}else{
		/* if the packet is not a MOD 8 packet, then we issue a
		 * Diagnostic packet with diagnostic code #40 (C1-B1)*/

		DEBUG_EVENT("%s: Error: Rx Invalid GFI 0x%X pkt: not Mod8 or Mod128\n",
					x25_link->name,frame->gfi);
		return 40;	
	}


	/* for an invalid bit setting */
	/* if the Q-bit is set, then the only valid packet type is 
	 * a Data packet (ISO 12.1.1) - for any other */
	/* packet, we issue a Diagnostic packet with diagnostic code #40 (C1-B1) */
	if (((frame->gfi & GFI_WITH_Q_BIT_SET) && (frame->type & BIT_0_SET))){
		DEBUG_EVENT("%s: Error: Rx Invalid pkt GFI 0x%X: Qbit on non data pkt \n",
				x25_link->name,frame->gfi);
		return 40;	/* discard the packet */
	}


	/* the D-bit of the GFI is set, so check the packet type */
	if(frame->gfi & GFI_WITH_D_BIT_SET) {				

		switch(frame->type) {
			
		/* the D-bit may only be set in Data, 
		 * Call Request and Call */
		case DATA_PKT:	
		case CALL_REQUEST_PKT:
		case CALL_ACCEPT_PKT:
			break;

		default:
			/* issue a Diagnostic packet with diagnostic 
			 * code #40 (C1-B1) */
			DEBUG_EVENT("%s: Error: Rx Invalid GFI 0x%X pkt: Dbit set on invalid pkt\n",
				x25_link->name,frame->gfi);
			return 40;	
		}
	}

	
	/* Non Zero Unassigned LCN: A=DIAG B=36 */
#if 0
	if (frame->lcn && x25_get_lcn_dev(x25_link,frame->lcn) == NULL){
		/* A=DIAG B=36 */
		printk(KERN_INFO "Unasigned LCN: A=DIAG D=36\n");
		return -EINVAL;
	}
#endif
	return 0;
}


static int x25_decode_pkt_type (x25_link_t *x25, struct sk_buff *skb, struct x25_frame *frame)
{
	WAN_ASSERT2((!x25),1);
	WAN_ASSERT2((!skb),1);
	WAN_ASSERT2((!frame),1);

	if (skb->len < LGTH_X25_PACKET_HEADER){
		
		frame->type = NO_SPECIFIC_PKT_TYPE;

		printk(KERN_INFO "Error: Packet type less than 1\n");
		/* handle an unsupported packet type */
		PTI_err_LCN_0_pkt_on_non_0_LCN (x25,UNIDENTIFIABLE_PKT,frame);
		return 1;
	}

	frame->type = skb->data[2];

	/* Data packets will be decoded later when we 
	 * found out for which LCN its destined for */
	if (!test_bit(0,&frame->type)){
		frame->type = DATA_PKT;
		return 0;
	}
	
	
	switch (frame->type & FLOW_CONTROL_MASK){

	case RR_PKT:
	case RNR_PKT:
	case REJ_PKT:
		frame->type = frame->type & FLOW_CONTROL_MASK;
		return 0;
	}

	
	/* Received Asynchronous Packet */

	skb_pull(skb,LGTH_X25_PACKET_HEADER);
		
	switch(frame->type) {

		/* these packet types include cause and diagnostic fields */
		case RESTART_REQUEST_PKT:
			printk(KERN_INFO "%s: X25 Event: Rx Restart Request: Cause 0x%02X, Diag 0x%02X\n",
					x25->name, skb->data[0],skb->data[1]); 
			x25_log_proc_event(x25,NULL,"Rx Restart Request",
			   		   0,skb->data[0],skb->data[1],NULL);

		case CLEAR_REQUEST_PKT:
		case RESET_REQUEST_PKT:
		case DIAGNOSTIC_PKT:
		case REGISTRATION_CONF_PKT:
			
			frame->cause = skb->data[0]; /* read the cause field */
			skb_pull(skb,1);
			/* if the packet includes a diagnostic field, then read it */
			if (skb->len > 0){
				frame->diag = skb->data[0];
				skb_pull(skb,1);
			}	
			break;
				
		case RESTART_CONF_PKT:
			x25_log_proc_event(x25,NULL,"Rx Restart Conf",
			   		   0,0,0,NULL);
			break;

		default:
			break;
	}
	
    	/* FIXME Update statistics */	
     #if 0 
	/* the asynchronous packet is an incoming packet */
	X25_history_misc_bits = HIST_PKT_RECEIVED;

	/* update our history table */
	update_X25_history_table(Rx_pkt_type);
     #endif

	/* Pass the Packet Up the next layer */
	return 0;
}

int x25_decode_data_ptk(x25_lcn_t *x25_lcn, struct sk_buff *skb, struct x25_frame *frame)
{
	
	WAN_ASSERT2((!x25_lcn),-EINVAL);
	WAN_ASSERT2((!skb),-EINVAL);
	WAN_ASSERT2((!frame),-EINVAL);

	
	if (x25_lcn->x25_link->cfg.mode & GFI_MOD_8_BIT_SET){

		if (!(frame->gfi & GFI_MOD_8_BIT_SET)){

			printk(KERN_INFO "Modulo Mode Mismatch !!! Clearing Call\n");

			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 
				 0, 
				 frame->lcn,
				 CLEAR_REQUEST_PKT,
				 0,
				 INVALID_Pr_COUNT,NULL,0);
			
			return -EINVAL;
		}

		switch (frame->type){

		case DATA_PKT:
			
			frame->mbit   = (skb->data[2] >> 4) & 0x01;
			frame->Pr_count = (skb->data[2] >> 5) & 0x07;
			frame->Ps_count = (skb->data[2] >> 1) & 0x07;
			break;
			
		case REJ_PKT:
		case RNR_PKT:
		case RR_PKT:
			frame->Pr_count = (skb->data[2] >> 5) & 0x07;
			break;

		default:
			/* Let all async packet throught */
			return -EINVAL;
		}

		skb_pull(skb,LGTH_X25_PACKET_HEADER);
	}else{
		
		if (!(frame->gfi & GFI_MOD_128_BIT_SET)){

			printk(KERN_INFO "Modulo Mode Mismatch !!! Clearing Call\n");

			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 
				 0, 
				 frame->lcn,
				 CLEAR_REQUEST_PKT,
				 0,
				 INVALID_Pr_COUNT,NULL,0);
			return -EINVAL;
		}
		
		if (skb->len < LGTH_X25_PACKET_HEADER + 1){

			printk(KERN_INFO "Packet too short MODULO 128\n");
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
			x25_send_control(x25_lcn->x25_link, 
				 0, 
				 atomic_read(&x25_lcn->lcn),
				 CLEAR_REQUEST_PKT,
				 0,UNIDENTIFIABLE_PKT,
				 NULL,0);
			return -EINVAL;
		}
	

		switch (frame->type){

		case DATA_PKT:

			frame->Ps_count = (skb->data[2] >> 1) & 0x7F;
			frame->Pr_count = (skb->data[3] >> 1) & 0x7F;
			frame->mbit   =  (skb->data[3] & 0x01);
			break;
			
		case REJ_PKT:
		case RNR_PKT:
		case RR_PKT:
			frame->Pr_count = (skb->data[3]>>1)&0x7F;
			break;

		default:
			return  -EINVAL;
		}
			
		skb_pull(skb,LGTH_X25_PACKET_HEADER+1);	
	}
	return 0;
}

int x25_lcn_sanity_check(x25_link_t *x25_link,x25_lcn_t *x25_lcn, struct x25_frame *frame)
{
	
	int len = frame->len-LGTH_X25_PACKET_HEADER;
	int cause=0;
	int diag=0;
	int err=0;

	if (len > x25_lcn->rx_pkt_size){
		err=-EINVAL;
		cause=0x05;
		diag=PKT_TOO_LONG;
	}else if (len < 0){
		err=-EINVAL;
		cause=0x05;
		diag=PKT_TOO_SHORT;
	}
	
	
	if (err){
		if (IS_PVC(x25_lcn)){
			lcn_state_change(x25_lcn,X25_ST_RESET_REQ_SNT_D2);
		
			x25_send_control(x25_link, 
			         0, 
				 frame->lcn,
				 RESET_REQUEST_PKT,
				 cause,diag,NULL,0);
		}else{
			lcn_state_change(x25_lcn,X25_ST_CLEAR_REQ_SNT_P6);
		
			x25_send_control(x25_link, 
			         0, 
				 frame->lcn,
				 CLEAR_REQUEST_PKT,
				 cause,diag,NULL,0);
		}	
		
	}
	return err;
}

/*
 *	Process an incoming X25 frame
 */
int x25_data_input(x25_link_t *x25_link, struct sk_buff *skb)
{
	struct x25_frame frame;
	struct net_device *dev=NULL;
	x25_lcn_t *x25_lcn=NULL;
	unsigned int old_state=-1;
	int err=-EINVAL;

	WAN_ASSERT2((!x25_link),-EINVAL);
	WAN_ASSERT2((!skb),-EINVAL);

	if (test_and_set_bit(RX_FSM_RUNNING,&x25_link->tq_working)){
		DEBUG_EVENT("%s: Critical SMP in Rx\n",x25_link->name);
		return err;
	}

	memset(&frame,0,sizeof(struct x25_frame));
	
	/* if packet structure is faulty, drop packet */
	if (x25_pkt_check_any(x25_link,skb,&frame) != X25_OK){
		goto end_data_input;
	}

	if (x25_decode_pkt_type(x25_link,skb,&frame) != X25_OK){
		goto end_data_input;
	}

	x25_update_stats(x25_link,frame.type,0);

	switch (x25_link->state) {
		
		case X25_ST_PKT_LVL_RDY_R1:

			if (frame.lcn)
				break;

			x25_state_r1_machine(x25_link, skb, &frame);
			goto end_data_input;
	
		case X25_ST_RESTART_REQ_SNT_R2:
			x25_state_r2_machine(x25_link, skb, &frame);
			goto end_data_input;

		case X25_ST_RESTART_REQ_REC_R3:
			x25_state_r3_machine(x25_link, skb, &frame);
			goto end_data_input;

		default:
			printk(KERN_INFO "Invallid X25 State %i\n",x25_link->state);
			goto end_data_input;
	}

	dev = x25_get_lcn_dev(x25_link,frame.lcn);
	if (dev == NULL){
		err=-EINVAL;
		x25_state_no_dev_machine(x25_link, skb, &frame);
		goto end_data_input;
	}
	
	x25_lcn = (x25_lcn_t*)dev->priv;
	old_state = x25_lcn->state;

	err=x25_lcn_sanity_check(x25_link,x25_lcn,&frame);
	if (err){
		goto end_data_input;
	}
	
	set_bit(LCN_RX, &x25_lcn->critical);

	if (test_bit(LCN_UNREGISTER,&x25_lcn->critical)){
		err=-EINVAL;
		goto end_data_input;
	}
	
	/* Update the incoming event */
//DEBUG FIXME
//possible race condition with other api_events
//and reading of api events
#if 0	
	if (frame.type >= CALL_REQUEST_PKT){
		x25_lcn->x25_api_event.hdr.qdm = (frame.gfi >> 1);
		if (frame.mbit){
			set_bit(0,&x25_lcn->x25_api_event.hdr.qdm);
		}else{
			clear_bit(0,&x25_lcn->x25_api_event.hdr.qdm);
		}
		x25_lcn->x25_api_event.hdr.cause = frame.cause;
		x25_lcn->x25_api_event.hdr.diagn = frame.diag;
		x25_lcn->x25_api_event.hdr.pktType = frame.type;
		x25_lcn->x25_api_event.hdr.length  = 0;
		x25_lcn->x25_api_event.hdr.result = 0;
		x25_lcn->x25_api_event.hdr.lcn = frame.lcn;
		x25_lcn->x25_api_event.hdr.mtu = x25_lcn->tx_pkt_size;
		x25_lcn->x25_api_event.hdr.mru = x25_lcn->rx_pkt_size;
	}
#endif

	if (IS_PVC(x25_lcn)){
	
		switch (x25_lcn->state){

			 /* Same as X25_ST_DATA_TRANSFER_P4 */
			case X25_ST_PKT_LVL_RDY_D1: 
			case X25_ST_PVC_OUT_OF_ORDER:
				err = x25_state_p4_machine(x25_lcn, skb, &frame);	
				break;

			case X25_ST_RESET_REQ_SNT_D2:
				x25_state_d2_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_RESET_REQ_REC_D3:
				x25_state_d3_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_INTERRUPT_SNT_I2:
				x25_state_i2_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_INTERRUPT_REC_I3:
				x25_state_i3_machine(x25_lcn, skb, &frame);
				break;
		}
		
	}else{
	
		switch (x25_lcn->state){

			case X25_ST_PKT_LVL_RDY_P1:
				x25_state_p1_machine(x25_lcn, skb, &frame);
				break;
			
			case X25_ST_CALL_REQ_SNT_P2:
				x25_state_p2_machine(x25_lcn, skb, &frame);
				break;
				
			case X25_ST_CALL_REQ_REC_P3:
				x25_state_p3_machine(x25_lcn, skb, &frame);
				break;

			//Note: I1 = D1 = P4	
			//case X25_ST_PKT_LVL_RDY_I1:
			//case X25_ST_PKT_LVL_RDY_D1:
			case X25_ST_DATA_TRANSFER_P4:
			case X25_ST_PVC_OUT_OF_ORDER:	
				err = x25_state_p4_machine(x25_lcn, skb, &frame);
				break;

			//case X25_ST_CALL_COLLISION_P5:
			//	x25_state_p5_machine(x25_lcn, skb, &frame);
			//	break;

			case X25_ST_CLEAR_REQ_SNT_P6:
				x25_state_p6_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_CLEAR_REQ_REC_P7:
				x25_state_p7_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_RESET_REQ_SNT_D2:
				x25_state_d2_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_RESET_REQ_REC_D3:
				x25_state_d3_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_INTERRUPT_SNT_I2:
				x25_state_i2_machine(x25_lcn, skb, &frame);
				break;

			case X25_ST_INTERRUPT_REC_I3:
				x25_state_i3_machine(x25_lcn, skb, &frame);
				break;
		}	
	}

end_data_input:
	if (err < 0){
		err=0;
		KFREE_SKB(skb);
	}

	if (x25_lcn){
		clear_bit(LCN_RX, &x25_lcn->critical);
	}

	if (dev){
		X25_PUT(dev);
	}

	clear_bit(RX_FSM_RUNNING,&x25_link->tq_working);

	return err;
}

